# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import os
import numpy as np
import h5py
import six
import csv

from .network import Network
from bmtk.builder.node import Node
from bmtk.builder.edge import Edge
from bmtk.utils import sonata

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    mpi_rank = comm.Get_rank()
    mpi_size = comm.Get_size()
    barrier = comm.barrier

except ImportError:
    mpi_rank = 0
    mpi_size = 1
    barrier = lambda: None


class DenseNetworkOrig(Network):
    def __init__(self, name, **network_props):
        super(DenseNetworkOrig, self).__init__(name, **network_props or {})

        self.__edges_types = {}
        self.__src_mapping = {}

        self.__networks = {}
        self.__node_count = 0
        self._nodes = []

        self.__edges_tables = []
        self._target_networks = {}

    def _initialize(self):
        self.__id_map = []
        self.__lookup = []

    def _add_nodes(self, nodes):
        self._nodes.extend(nodes)
        self._nnodes = len(self._nodes)

        """
        id_label = 'node_id' if 'node_id' in nodes[0].keys() else 'id'

        start_idx = len(self.__id_map)  #
        self.__id_map += [n[id_label] for n in nodes]
        self.__nodes += [(interal_id, nodes[node_idx])
                         for node_idx, interal_id in enumerate(xrange(start_idx, len(self.__id_map)))]

        assert(len(self.__id_map) == len(self.__nodes))
        """

    def edges_table(self):
        return self.__edges_tables

    def _save_nodes(self, nodes_file_name):
        if not self._nodes_built:
            self._build_nodes()

        # save the node_types file
        # TODO: how do we add attributes to the h5
        group_indx = 0
        groups_lookup = {}
        group_indicies = {}
        group_props = {}
        for ns in self._node_sets:
            if ns.params_hash in groups_lookup:
                continue
            else:
                groups_lookup[ns.params_hash] = group_indx
                group_indicies[group_indx] = 0
                group_props[group_indx] = {k: [] for k in ns.params_keys if k != 'node_id'}
                group_indx += 1

        node_gid_table = np.zeros(self._nnodes)  # todo: set dtypes
        node_type_id_table = np.zeros(self._nnodes)
        node_group_table = np.zeros(self._nnodes)
        node_group_index_tables = np.zeros(self._nnodes)

        for i, node in enumerate(self.nodes()):
            node_gid_table[i] = node.node_id
            node_type_id_table[i] = node.node_type_id
            group_id = groups_lookup[node.params_hash]
            node_group_table[i] = group_id
            node_group_index_tables[i] = group_indicies[group_id]
            group_indicies[group_id] += 1

            group_dict = group_props[group_id]
            for key, prop_ds in group_dict.items():
                prop_ds.append(node.params[key])

        # TODO: open in append mode
        if mpi_rank == 0:
            with h5py.File(nodes_file_name, 'w') as hf:
                # Add magic and version attribute
                add_hdf5_attrs(hf)

                pop_grp = hf.create_group('/nodes/{}'.format(self.name))
                pop_grp.create_dataset('node_id', data=node_gid_table, dtype='uint64')
                pop_grp.create_dataset('node_type_id', data=node_type_id_table, dtype='uint64')
                pop_grp.create_dataset('node_group_id', data=node_group_table, dtype='uint32')
                pop_grp.create_dataset('node_group_index', data=node_group_index_tables, dtype='uint64')

                for grp_id, props in group_props.items():
                    model_grp = pop_grp.create_group('{}'.format(grp_id))

                    for key, dataset in props.items():
                        # ds_path = 'nodes/{}/{}'.format(grp_id, key)
                        try:
                            model_grp.create_dataset(key, data=dataset)
                        except TypeError:
                            str_list = [str(d) for d in dataset]
                            hf.create_dataset(key, data=str_list)
        barrier()

    def nodes_iter(self, node_ids=None):
        if node_ids is not None:
            return [n for n in self._nodes if n.node_id in node_ids]
        else:
            return self._nodes

    def _process_nodepool(self, nodepool):
        return nodepool

    def import_nodes(self, nodes_file_name, node_types_file_name, population=None):
        sonata_file = sonata.File(data_files=nodes_file_name, data_type_files=node_types_file_name)
        if sonata_file.nodes is None:
            raise Exception('nodes file {} does not have any nodes.'.format(nodes_file_name))

        populations = sonata_file.nodes.populations
        if len(populations) == 1:
            node_pop = populations[0]
        elif population is None:
            raise Exception('The nodes file {} contains multiple populations.'.format(nodes_file_name) +
                            'Please specify population parameter.')
        else:
            for pop in populations:
                if pop.name == population:
                    node_pop = pop
                    break
            else:
                raise Exception('Nodes file {} does not contain population {}.'.format(nodes_file_name, population))

        for node_type_props in node_pop.node_types_table:
            self._add_node_type(node_type_props)

        for node in node_pop:
            self._node_id_gen.remove_id(node.node_id)
            self._nodes.append(Node(node.node_id, node.group_props, node.node_type_properties))

    def _add_edges(self, connection_map, i):
        syn_table = self.EdgeTable(connection_map)
        connections = connection_map.connection_itr()
        for con in connections:
            if con[2] is not None:
                syn_table[con[0], con[1]] = con[2]

        target_net = connection_map.target_nodes
        self._target_networks[target_net.network_name] = target_net.network

        nsyns = np.sum(syn_table.nsyn_table)
        self._nedges += int(nsyns)
        edge_table = {'syn_table': syn_table,
                      'nsyns': nsyns,
                      'edge_types': connection_map.edge_type_properties,
                      'edge_type_id': connection_map.edge_type_properties['edge_type_id'],
                      'source_network': connection_map.source_nodes.network_name,
                      'target_network': connection_map.target_nodes.network_name,
                      'params': {},
                      'params_dtypes': {},
                      'source_query': connection_map.source_nodes.filter_str,
                      'target_query': connection_map.target_nodes.filter_str}


        for param in connection_map.params:
            rule = param.rule
            param_names = param.names
            edge_table['params_dtypes'].update(param.dtypes)
            if isinstance(param_names, list) or isinstance(param_names, tuple):
                tmp_tables = [self.PropertyTable(nsyns, param.dtypes[n]) for n in param_names]
                # tmp_tables = [self.PropertyTable(nsyns) for _ in range(len(param_names))]
                for source in connection_map.source_nodes:
                    src_node_id = source.node_id
                    for target in connection_map.target_nodes:
                        trg_node_id = target.node_id  # TODO: pull this out and put in it's own list
                        for _ in range(syn_table[src_node_id, trg_node_id]):
                            pvals = rule(source, target)
                            for i in range(len(param_names)):
                                tmp_tables[i][src_node_id, trg_node_id] = pvals[i]

                for i, name in enumerate(param_names):
                    # TODO: I think a copy constructor might get called, move this out.
                    edge_table['params'][name] = tmp_tables[i]

            else:
                pt = self.PropertyTable(np.sum(nsyns), dtype=param.dtypes[param_names])
                for source in connection_map.source_nodes:
                    src_node_id = source.node_id
                    for target in connection_map.target_nodes:
                        trg_node_id = target.node_id  # TODO: pull this out and put in it's own list
                        #print('{}, {}: {}'.format(src_node_id, trg_node_id, edge_table[src_node_id, trg_node_id]))
                        for _ in range(syn_table[src_node_id, trg_node_id]):
                            pt[src_node_id, trg_node_id] = rule(source, target)
                edge_table['params'][param_names] = pt

        self.__edges_tables.append(edge_table)


    def _save_gap_junctions(self, gj_file_name):
        source_ids = []
        target_ids = []
        src_gap_ids = []
        trg_gap_ids = []

        for et in self.__edges_tables:
            try:
                is_gap = et['edge_types']['is_gap_junction']
            except:
                continue
            if is_gap:
                if et['source_network'] != et['target_network']:
                    raise Exception("All gap junctions must be two cells in the same network builder.")
                
                table = et['syn_table']
                junc_table = table.nsyn_table
                locs = np.where(junc_table > 0)
                for i in range(len(locs[0])):
                    source_ids.append(table.source_ids[locs[0][i]])
                    target_ids.append(table.target_ids[locs[1][i]])
                    src_gap_ids.append(self._gj_id_gen.next())
                    trg_gap_ids.append(self._gj_id_gen.next())
            else:
                continue
        
        if len(source_ids) > 0:
            with h5py.File(gj_file_name, 'w') as f:
                add_hdf5_attrs(f)
                f.create_dataset('source_ids', data=np.array(source_ids))
                f.create_dataset('target_ids', data=np.array(target_ids))
                f.create_dataset('src_gap_ids', data=np.array(src_gap_ids))
                f.create_dataset('trg_gap_ids', data=np.array(trg_gap_ids))


    def _save_edges(self, edges_file_name, src_network, trg_network, name=None):
        groups = {}
        group_dtypes = {}  # TODO: this should be stored in PropertyTable
        grp_id_itr = 0
        groups_lookup = {}
        total_syns = 0

        matching_edge_tables = [et for et in self.__edges_tables
                                if et['source_network'] == src_network and et['target_network'] == trg_network]

        for ets in matching_edge_tables:
            params_hash = str(ets['params'].keys())
            group_id = groups_lookup.get(params_hash, None)
            if group_id is None:
                group_id = grp_id_itr
                groups_lookup[params_hash] = group_id
                grp_id_itr += 1

            ets['group_id'] = group_id
            groups[group_id] = {}
            group_dtypes[group_id] = ets['params_dtypes']
            for param_name in ets['params'].keys():
                groups[group_id][param_name] = []

            total_syns += int(ets['nsyns'])

        group_index_itrs = [0 for _ in range(grp_id_itr)]
        trg_gids = np.zeros(total_syns)  # set dtype to uint64
        src_gids = np.zeros(total_syns)
        edge_groups = np.zeros(total_syns)  # dtype uint16 or uint8
        edge_group_index = np.zeros(total_syns)  # uint32
        edge_type_ids = np.zeros(total_syns)  # uint32

        # TODO: Another potential issue if node-ids don't start with 0
        index_ptrs = np.zeros(len(self._target_networks[trg_network].nodes()) + 1)
        #index_ptrs = np.zeros(len(self._nodes)+1)  # TODO: issue when target nodes come from another network
        index_ptr_itr = 0

        gid_indx = 0
        for trg_node in self._target_networks[trg_network].nodes():
            index_ptrs[index_ptr_itr] = gid_indx
            index_ptr_itr += 1

            for ets in matching_edge_tables:
                edge_group_id = ets['group_id']
                group_table = groups[edge_group_id]

                syn_table = ets['syn_table']
                if syn_table.has_target(trg_node.node_id):
                    if ets['params']:
                        for src_id, nsyns in syn_table.trg_itr(trg_node.node_id):
                            # Add on to the edges index
                            indx_end = gid_indx+nsyns
                            while gid_indx < indx_end:
                                trg_gids[gid_indx] = trg_node.node_id
                                src_gids[gid_indx] = src_id
                                edge_type_ids[gid_indx] = ets['edge_type_id']
                                edge_groups[gid_indx] = edge_group_id
                                edge_group_index[gid_indx] = group_index_itrs[edge_group_id]
                                group_index_itrs[edge_group_id] += 1
                                gid_indx += 1

                            for param_name, param_table in ets['params'].items():
                                param_vals = group_table[param_name]
                                for val in param_table.itr_vals(src_id, trg_node.node_id):
                                    param_vals.append(val)

                    else:
                        # If no properties just print nsyns table.
                        if 'nsyns' not in group_table:
                            group_table['nsyns'] = []
                        group_dtypes[edge_group_id]['nsyns'] = 'uint16'
                        for src_id, nsyns in syn_table.trg_itr(trg_node.node_id):
                            trg_gids[gid_indx] = trg_node.node_id
                            src_gids[gid_indx] = src_id
                            edge_type_ids[gid_indx] = ets['edge_type_id']
                            edge_groups[gid_indx] = edge_group_id
                            edge_group_index[gid_indx] = group_index_itrs[edge_group_id]
                            group_index_itrs[edge_group_id] += 1
                            gid_indx += 1

                            group_table['nsyns'].append(nsyns)

        trg_gids = trg_gids[:gid_indx]
        src_gids = src_gids[:gid_indx]
        edge_groups = edge_groups[:gid_indx]
        edge_group_index = edge_group_index[:gid_indx]
        edge_type_ids = edge_type_ids[:gid_indx]

        pop_name = '{}_to_{}'.format(src_network, trg_network) if name is None else name

        index_ptrs[index_ptr_itr] = gid_indx
        with h5py.File(edges_file_name, 'w') as hf:
            add_hdf5_attrs(hf)
            pop_grp = hf.create_group('/edges/{}'.format(pop_name))
            pop_grp.create_dataset('target_node_id', data=trg_gids, dtype='uint64')
            pop_grp['target_node_id'].attrs['node_population'] = trg_network
            pop_grp.create_dataset('source_node_id', data=src_gids, dtype='uint64')
            pop_grp['source_node_id'].attrs['node_population'] = src_network

            pop_grp.create_dataset('edge_group_id', data=edge_groups, dtype='uint16')
            pop_grp.create_dataset('edge_group_index', data=edge_group_index, dtype='uint32')
            pop_grp.create_dataset('edge_type_id', data=edge_type_ids, dtype='uint32')
            # pop_grp.create_dataset('edges/index_pointer', data=index_ptrs, dtype='uint32')

            for group_id, params_dict in groups.items():
                model_grp = pop_grp.create_group(str(group_id))
                for params_key, params_vals in params_dict.items():
                    #group_path = 'edges/{}/{}'.format(group_id, params_key)
                    dtype = group_dtypes[group_id][params_key]
                    if dtype is not None:
                        model_grp.create_dataset(params_key, data=list(params_vals), dtype=dtype)
                    else:
                        model_grp.create_dataset(params_key, data=list(params_vals))

            self._create_index(pop_grp['target_node_id'], pop_grp, index_type='target')
            self._create_index(pop_grp['source_node_id'], pop_grp, index_type='source')

    def _create_index(self, node_ids_ds, output_grp, index_type='target'):
        if index_type == 'target':
            edge_nodes = np.array(node_ids_ds, dtype=np.int64)
            output_grp = output_grp.create_group('indices/target_to_source')
        elif index_type == 'source':
            edge_nodes = np.array(node_ids_ds, dtype=np.int64)
            output_grp = output_grp.create_group('indices/source_to_target')

        edge_nodes = np.append(edge_nodes, [-1])
        n_targets = np.max(edge_nodes)
        ranges_list = [[] for _ in six.moves.range(n_targets + 1)]

        n_ranges = 0
        begin_index = 0
        cur_trg = edge_nodes[begin_index]
        for end_index, trg_gid in enumerate(edge_nodes):
            if cur_trg != trg_gid:
                ranges_list[cur_trg].append((begin_index, end_index))
                cur_trg = int(trg_gid)
                begin_index = end_index
                n_ranges += 1

        node_id_to_range = np.zeros((n_targets + 1, 2))
        range_to_edge_id = np.zeros((n_ranges, 2))
        range_index = 0
        for node_index, trg_ranges in enumerate(ranges_list):
            if len(trg_ranges) > 0:
                node_id_to_range[node_index, 0] = range_index
                for r in trg_ranges:
                    range_to_edge_id[range_index, :] = r
                    range_index += 1
                node_id_to_range[node_index, 1] = range_index

        output_grp.create_dataset('range_to_edge_id', data=range_to_edge_id, dtype='uint64')
        output_grp.create_dataset('node_id_to_range', data=node_id_to_range, dtype='uint64')

    def _clear(self):
        self._nedges = 0
        self._nnodes = 0

    def edges_iter(self, trg_gids, src_network=None, trg_network=None):
        matching_edge_tables = self.__edges_tables
        if trg_network is not None:
            matching_edge_tables = [et for et in self.__edges_tables if et['target_network'] == trg_network]

        if src_network is not None:
            matching_edge_tables = [et for et in matching_edge_tables if et['source_network'] == src_network]

        for trg_gid in trg_gids:
            for ets in matching_edge_tables:
                syn_table = ets['syn_table']
                if syn_table.has_target(trg_gid):
                    for src_id, nsyns in syn_table.trg_itr(trg_gid):
                        if ets['params']:
                            synapses = [{} for _ in range(nsyns)]
                            for param_name, param_table in ets['params'].items():
                                for i, val in enumerate(param_table[src_id, trg_gid]):
                                    synapses[i][param_name] = val
                            for syn_prop in synapses:
                                yield Edge(src_gid=src_id, trg_gid=trg_gid, edge_type_props=ets['edge_types'],
                                           syn_props=syn_prop)
                        else:
                            yield Edge(src_gid=src_id, trg_gid=trg_gid, edge_type_props=ets['edge_types'],
                                       syn_props={'nsyns': nsyns})

    @property
    def nnodes(self):
        if not self.nodes_built:
            return 0
        return self._nnodes

    @property
    def nedges(self):
        return self._nedges

    class EdgeTable(object):
        def __init__(self, connection_map):
            # TODO: save column and row lengths
            # Create maps between source_node gids and their row in the matrix.
            self.__idx2src = [n.node_id for n in connection_map.source_nodes]
            self.__src2idx = {node_id: i for i, node_id in enumerate(self.__idx2src)}

            # Create maps betwee target_node gids and their column in the matrix
            self.__idx2trg = [n.node_id for n in connection_map.target_nodes]
            self.__trg2idx = {node_id: i for i, node_id in enumerate(self.__idx2trg)}

            self._nsyn_table = np.zeros((len(self.__idx2src), len(self.__idx2trg)), dtype=np.uint8)

        def __getitem__(self, item):
            # TODO: make sure matrix is column oriented, or swithc trg and srcs.
            indexed_pair = (self.__src2idx[item[0]], self.__trg2idx[item[1]])
            return self._nsyn_table[indexed_pair]

        def __setitem__(self, key, value):
            assert(len(key) == 2)
            indexed_pair = (self.__src2idx[key[0]], self.__trg2idx[key[1]])
            self._nsyn_table[indexed_pair] = value

        def has_target(self, node_id):
            return node_id in self.__trg2idx

        @property
        def nsyn_table(self):
            return self._nsyn_table

        @property
        def target_ids(self):
            return self.__idx2trg

        @property
        def source_ids(self):
            return self.__idx2src

        def trg_itr(self, trg_id):
            trg_i = self.__trg2idx[trg_id]
            for src_j, src_id in enumerate(self.__idx2src):
                nsyns = self._nsyn_table[src_j, trg_i]
                if nsyns:
                    yield src_id, nsyns

    class PropertyTable(object):
        # TODO: add support for strings
        def __init__(self, nvalues, dtype):
            self._prop_array = np.zeros(nvalues, dtype=dtype)
            # self._prop_table = np.zeros((nvalues, 1))  # TODO: set dtype
            self._index = np.zeros((nvalues, 2), dtype=np.uint32)
            self._itr_index = 0

        def itr_vals(self, src_id, trg_id):
            indicies = np.where((self._index[:, 0] == src_id) & (self._index[:, 1] == trg_id))
            for val in self._prop_array[indicies]:
                yield val

        def __setitem__(self, key, value):
            self._index[self._itr_index, 0] = key[0]  # src_node_id
            self._index[self._itr_index, 1] = key[1]  # trg_node_id
            self._prop_array[self._itr_index] = value
            self._itr_index += 1

        def __getitem__(self, item):
            indicies = np.where((self._index[:, 0] == item[0]) & (self._index[:, 1] == item[1]))
            return self._prop_array[indicies]


def add_hdf5_attrs(hdf5_handle):
    # TODO: move this as a utility function
    hdf5_handle['/'].attrs['magic'] = np.uint32(0x0A7A)
    hdf5_handle['/'].attrs['version'] = [np.uint32(0), np.uint32(1)]



import numpy as np
import scipy.stats as sps

from .make_cell_list import multi_cell_random_generator, make_single_unit_cell_list, make_on_off_cell_list


def make_lattice_unit(lattice_unit_center=None):
    cell_list = []
    tON_cell_list = make_tON_cell_list(lattice_unit_center)
    tOFF_cell_list = make_tOFF_cell_list(lattice_unit_center)
    sON_cell_list = make_sON_cell_list(lattice_unit_center)
    sOFF_cell_list = make_sOFF_cell_list(lattice_unit_center)
    overlap_onoff_cell_list = make_overlapping_onoff_cell_list(lattice_unit_center)
    separate_onoff_cell_list = make_separate_onoff_cell_list(lattice_unit_center)
    
    cell_list = tON_cell_list + tOFF_cell_list + sON_cell_list + sOFF_cell_list + overlap_onoff_cell_list + separate_onoff_cell_list 
    
    return cell_list
    

def make_tON_cell_list(lattice_unit_center):
    tON_cell_list = []
    
    single_unit_cell_config = {}
    single_unit_cell_config['lattice_unit_center']=lattice_unit_center
    single_unit_cell_config['width'] = 5.
    sz = [3,6,9]
    ncells = [5,3,2]
    amp_dist = sps.rv_discrete(values=([20,25], [.5,.5]))
    kpeaks_dist =  sps.multivariate_normal(mean=[15., 35.], cov=[[5.0, 0], [0, 5]])
    wts = (4.,-2.5)
    delays = (0.,0.)
    single_unit_cell_config['amplitude'] = amp_dist
    single_unit_cell_config['kpeaks'] = kpeaks_dist
    single_unit_cell_config['weights'] = wts
    single_unit_cell_config['delays'] = delays
    for num_cells, sig in zip(ncells,sz):
        single_unit_cell_config['number_of_cells'] = num_cells
        single_unit_cell_config['sigma'] = (sig,sig)
        tON_cell_list += multi_cell_random_generator(make_single_unit_cell_list, **single_unit_cell_config)
    
    return tON_cell_list

def make_tOFF_cell_list(lattice_unit_center):
    tOFF_cell_list = []
    
    single_unit_cell_config = {}
    single_unit_cell_config['lattice_unit_center']=lattice_unit_center
    single_unit_cell_config['width'] = 5.
    sz = [3,6,9]
    ncells = [10,5,5]
    amp_dist = sps.rv_discrete(values=([-20,-25], [.5,.5]))
#     kpeaks_dist =  sps.multivariate_normal(mean=[40., 80.], cov=[[5.0, 0], [0, 5]])
#     wts = (.4,-.2)
    kpeaks_dist =  sps.multivariate_normal(mean=[15., 35.], cov=[[5.0, 0], [0, 5]])
    wts = (4.,-2.5)
    delays = (0.,0.)
    single_unit_cell_config['amplitude'] = amp_dist
    single_unit_cell_config['kpeaks'] = kpeaks_dist
    single_unit_cell_config['weights'] = wts
    single_unit_cell_config['delays'] = delays
    for num_cells, sig in zip(ncells,sz):
        single_unit_cell_config['number_of_cells'] = num_cells
        single_unit_cell_config['sigma'] = (sig,sig)
        tOFF_cell_list += multi_cell_random_generator(make_single_unit_cell_list, **single_unit_cell_config) 
    
    #print len(tOFF_cell_list)        
    return tOFF_cell_list

def make_sON_cell_list(lattice_unit_center):
    sON_cell_list = []
    
    single_unit_cell_config = {}
    single_unit_cell_config['lattice_unit_center']=lattice_unit_center
    single_unit_cell_config['width'] = 5.
    sz = [3,6,9]
    ncells = [5,3,2]
    amp_dist = sps.rv_discrete(values=([20,25], [.5, .5]))
    kpeaks_dist =  sps.multivariate_normal(mean=[80., 120.], cov=[[5.0, 0], [0, 5]])
    wts = (4.,-.85)
    delays = (0.,0.)
    single_unit_cell_config['amplitude'] = amp_dist
    single_unit_cell_config['kpeaks'] = kpeaks_dist
    single_unit_cell_config['weights'] = wts
    single_unit_cell_config['delays'] = delays
    for num_cells, sig in zip(ncells,sz):
        single_unit_cell_config['number_of_cells'] = num_cells
        single_unit_cell_config['sigma'] = (sig,sig)
        sON_cell_list += multi_cell_random_generator(make_single_unit_cell_list, **single_unit_cell_config) 
    
    return sON_cell_list


def make_sOFF_cell_list(lattice_unit_center):
    sOFF_cell_list = []
    
    single_unit_cell_config = {}
    single_unit_cell_config['lattice_unit_center']=lattice_unit_center
    single_unit_cell_config['width'] = 5.
    sz = [3,6,9]
    ncells = [10,5,5]
    amp_dist = sps.rv_discrete(values=([-20,-25], [.5,.5]))
    kpeaks_dist =  sps.multivariate_normal(mean=[80., 120.], cov=[[5.0, 0], [0, 5]])
    wts = (4.,-.85)
    delays = (0.,0.)
    single_unit_cell_config['amplitude'] = amp_dist
    single_unit_cell_config['kpeaks'] = kpeaks_dist
    single_unit_cell_config['weights'] = wts
    single_unit_cell_config['delays'] = delays
    for num_cells, sig in zip(ncells,sz):
        single_unit_cell_config['number_of_cells'] = num_cells
        single_unit_cell_config['sigma'] = (sig,sig)
        sOFF_cell_list += multi_cell_random_generator(make_single_unit_cell_list, **single_unit_cell_config) 
    
    return sOFF_cell_list


def make_overlapping_onoff_cell_list(lattice_unit_center):
    overlap_onoff_cell_list = []
    
    two_unit_cell_config = {}
    two_unit_cell_config['lattice_unit_center']=lattice_unit_center
    two_unit_cell_config['width']=5.
    
    ncells = 4
    sz = 9
    ang_dist = sps.rv_discrete(values=(np.arange(0,180,45), 1./ncells*np.ones(ncells)))
    amp_on_dist = sps.rv_discrete(values=([20,25], [.5,.5]))
    amp_off_dist = sps.rv_discrete(values=([-20,-25], [.5,.5]))
    kpeak_on_dist =  sps.multivariate_normal(mean=[15., 35.], cov=[[5.0, 0], [0, 5]])
    kpeak_off_dist =  sps.multivariate_normal(mean=[20., 40.], cov=[[5.0, 0], [0, 5]])
    wts_on = wts_off = (4.,-2.5)
    delays_on = delays_off = (0.,0.)
    subfield_sep = 2.
    
    two_unit_cell_config['number_of_cells'] = ncells
    two_unit_cell_config['ang'] = ang_dist
    two_unit_cell_config['amplitude_on'] = amp_on_dist
    two_unit_cell_config['amplitude_off'] = amp_off_dist
    two_unit_cell_config['kpeaks_on'] = kpeak_on_dist
    two_unit_cell_config['kpeaks_off'] = kpeak_off_dist
    two_unit_cell_config['weights_on'] = wts_on
    two_unit_cell_config['weights_off'] = wts_off
    two_unit_cell_config['sigma_on'] = (sz,sz)
    two_unit_cell_config['sigma_off'] = (sz,sz)
    two_unit_cell_config['subfield_separation'] = subfield_sep
    two_unit_cell_config['dominant_subunit']='on'
    two_unit_cell_config['delays_on']=delays_on
    two_unit_cell_config['delays_off']=delays_off
    
    overlap_onoff_cell_list += multi_cell_random_generator(make_on_off_cell_list, **two_unit_cell_config)
    
    return overlap_onoff_cell_list


def make_separate_onoff_cell_list(lattice_unit_center):
    separate_onoff_cell_list = []
    
    two_unit_cell_config = {}
    two_unit_cell_config['lattice_unit_center']=lattice_unit_center
    two_unit_cell_config['width']=5.
    
    ncells = 8
    sz = 6
    ang_dist = np.arange(0,360,45)
    subfield_sep = 4.

    kpeak_dom_dist =  sps.multivariate_normal(mean=[15., 35.], cov=[[5.0, 0], [0, 5]])
    kpeak_nondom_dist =  sps.multivariate_normal(mean=[80., 120.], cov=[[5.0, 0], [0, 5]])
    wts_dom = (4.,-2.5)
    wts_nondom = (4,-.85)
    delays_dom = delays_nondom = (0.,0.)    
    
    two_unit_cell_config['number_of_cells'] = ncells
    two_unit_cell_config['ang'] = ang_dist
    two_unit_cell_config['sigma_on'] = (sz,sz)
    two_unit_cell_config['sigma_off'] = (sz,sz)
    two_unit_cell_config['subfield_separation'] = subfield_sep
    
    #On-dominant
    dom_subunit = 'on'
    if dom_subunit=='on':
        two_unit_cell_config['dominant_subunit'] = dom_subunit
        amp_dom_dist = sps.rv_discrete(values=([20,25], [.5,.5]))
        amp_nondom_dist = sps.rv_discrete(values=([-10,-15], [.5,.5]))
        two_unit_cell_config['amplitude_on'] = amp_dom_dist
        two_unit_cell_config['amplitude_off'] = amp_nondom_dist
        two_unit_cell_config['kpeaks_on'] = kpeak_dom_dist
        two_unit_cell_config['kpeaks_off'] = kpeak_nondom_dist
        two_unit_cell_config['weights_on'] = wts_dom
        two_unit_cell_config['weights_off'] = wts_nondom
        two_unit_cell_config['delays_on'] = delays_dom
        two_unit_cell_config['delays_off'] = delays_nondom
        separate_onoff_cell_list += multi_cell_random_generator(make_on_off_cell_list, **two_unit_cell_config)
        
    #Off-dominant
    dom_subunit = 'off'
    if dom_subunit=='off':
        two_unit_cell_config['dominant_subunit'] = dom_subunit
        amp_dom_dist = sps.rv_discrete(values=([-20,-25], [.5,.5]))
        amp_nondom_dist = sps.rv_discrete(values=([10,15], [.5,.5]))
        two_unit_cell_config['amplitude_off'] = amp_dom_dist
        two_unit_cell_config['amplitude_on'] = amp_nondom_dist
        two_unit_cell_config['kpeaks_off'] = kpeak_dom_dist
        two_unit_cell_config['kpeaks_on'] = kpeak_nondom_dist
        two_unit_cell_config['weights_off'] = wts_dom
        two_unit_cell_config['weights_on'] = wts_nondom
        two_unit_cell_config['delays_off'] = delays_dom
        two_unit_cell_config['delays_on'] = delays_nondom
        separate_onoff_cell_list += multi_cell_random_generator(make_on_off_cell_list, **two_unit_cell_config)
    
    #print len(separate_onoff_cell_list)    
    return separate_onoff_cell_list

if __name__ == "__main__":
    lattice_unit_center = (40,30)
    lattice_cell_list = make_lattice_unit(lattice_unit_center)
    print(len(lattice_cell_list))
        


# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import os
import json

from neuron import h

from bmtk.simulator.core.simulation_config import SimulationConfig
from bmtk.simulator.bionet.io_tools import io
from . import nrn


pc = h.ParallelContext()    # object to access MPI methods
MPI_Rank = int(pc.id())


class Config(SimulationConfig):
    def create_output_dir(self):
        io.setup_output_dir(self.output_dir, self.log_file)

    def load_nrn_modules(self):
        nrn.load_neuron_modules(self.mechanisms_dir, self.templates_dir)

    def build_env(self):
        self.io = io
        self._set_logging()
        self.create_output_dir()
        self.copy_to_output()
        if io.mpi_size > 1:
            # A friendly message requested by fb
            io.log_info('Running NEURON with mpi ({} cores).'.format(io.mpi_size))

        pc.barrier()
        self.load_nrn_modules()



from .swc_reader import SWCReader
from .swc_reader import get_swc


def rand_syn_locations(src, trg, sections=('soma', 'apical', 'basal'), distance_range=(0.0, 1.0e20),
                       morphology_dir='./components/morphologies'):
    trg_swc = get_swc(trg, morphology_dir=morphology_dir, use_cache=True)

    sec_ids, seg_xs = trg_swc.choose_sections(sections, distance_range, n_sections=1)
    sec_id, seg_x = sec_ids[0], seg_xs[0]
    swc_id, swc_dist = trg_swc.get_swc_id(sec_id, seg_x)
    # coords = trg_swc.get_coords(sec_id, seg_x)

    return [sec_id, seg_x, swc_id, swc_dist]  # coords[0], coords[y], coords[z]



# Copyright 2020. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import csv
import pandas as pd
import numpy as np
import h5py
import six
from collections import defaultdict
import warnings

from .spike_trains_api import SpikeTrainsReadOnlyAPI
from .core import SortOrder, csv_headers, col_population, col_timestamps, col_node_ids, pop_na


GRP_spikes_root = 'spikes'
DATASET_timestamps = 'timestamps'
DATASET_node_ids = 'node_ids'


sorting_attrs = {
    'time': SortOrder.by_time,
    'by_time': SortOrder.by_time,
    'id': SortOrder.by_id,
    'by_id': SortOrder.by_id,
    'none': SortOrder.none,
    'unknown': SortOrder.unknown
}


def load_sonata_file(path, version=None, **kwargs):
    """Loads a Sonata file reader, making sure it matches the correct version.

    :param path:
    :param version:
    :param kwargs:
    :return:
    """
    try:
        with h5py.File(path, 'r') as h5:
            spikes_root = h5[GRP_spikes_root]
            for name, h5_obj in spikes_root.items():
                if isinstance(h5_obj, h5py.Group):
                    # In case there exists a population subgroup
                    return SonataSTReader(path, **kwargs)
    except Exception:
        pass

    try:
        with h5py.File(path, 'r') as h5:
            spikes_root = h5[GRP_spikes_root]
            if 'gids' in spikes_root and 'timestamps' in spikes_root:
                return SonataOldReader(path, **kwargs)
    except Exception:
        pass

    try:
        with h5py.File(path, 'r') as h5:
            if '/spikes' in h5:
                return EmptySonataReader(path, **kwargs)
    except Exception:
        pass

    raise Exception('Could not open file {}, does not contain SONATA spike-trains'.format(path))


def to_list(v):
    if v is not None and np.isscalar(v):
        return [v]
    else:
        return v


class SonataSTReader(SpikeTrainsReadOnlyAPI):
    def __init__(self, path, **kwargs):
        self._path = path
        self._h5_handle = h5py.File(self._path, 'r')
        self._DATASET_node_ids = 'node_ids'
        self._n_spikes = None
        # TODO: Create a function for looking up population and can return errors if more than one
        self._default_pop = None
        # self._node_list = None

        self._indexed = False
        self._index_nids = {}

        if GRP_spikes_root not in self._h5_handle:
            raise Exception('Could not find /{} root'.format(GRP_spikes_root))
        else:
            self._spikes_root = self._h5_handle[GRP_spikes_root]

        if 'population' in kwargs:
            pop_filter = to_list(kwargs['population'])
        elif 'populations' in kwargs:
            pop_filter = to_list(kwargs['populations'])
        else:
            pop_filter = None

        # get a map of 'pop_name' -> pop_group
        self._population_map = {}
        for name, h5_obj in self._h5_handle[GRP_spikes_root].items():
            if isinstance(h5_obj, h5py.Group):
                if pop_filter is not None and name not in pop_filter:
                    continue

                if 'node_ids' not in h5_obj or 'timestamps' not in h5_obj:
                    warnings.warn('population {} in {} is missing spikes, skipping.'.format(name, path))
                    continue

                self._population_map[name] = h5_obj

        if not self._population_map:
            # In old version of the sonata standard there was no 'population' subgroup. For backwards compatability
            # use a default dictionary
            # TODO: Remove so we only have to support latest version of SONATA
            self._population_map = defaultdict(lambda: self._h5_handle[GRP_spikes_root])
            self._population_map[pop_na] = self._h5_handle[GRP_spikes_root]
            self._DATASET_node_ids = 'gids'

        self._default_pop = kwargs.get('default_population', list(self._population_map.keys())[0])

        self._population_sorting_map = {}
        for pop_name, pop_grp in self._population_map.items():
            if 'sorting' in pop_grp[self._DATASET_node_ids].attrs.keys():
                # Found a few existing sonata files put the 'sorting' attribute in the node_ids dataset, remove later
                attr_str = pop_grp[self._DATASET_node_ids].attrs['sorting']
                sort_order = sorting_attrs.get(attr_str, SortOrder.unknown)

            elif 'sorting' in pop_grp.attrs.keys():
                attr_str = pop_grp.attrs['sorting']
                sort_order = sorting_attrs.get(attr_str, SortOrder.unknown)

            else:
                sort_order = SortOrder.unknown

            self._population_sorting_map[pop_name] = sort_order

        # TODO: Add option to skip building indices
        self._build_node_index()

        # units are not instrinsic to a csv file, but allow users to pass it in if they know
        self._units_maps = {}
        for pop_name, pop_grp in self._population_map.items():
            if 'units' in pop_grp['timestamps'].attrs:
                pop_units = pop_grp['timestamps'].attrs['units']
            elif 'units' in kwargs:
                pop_units = kwargs['units']
            else:
                pop_units = 'ms'

            self._units_maps[pop_name] = pop_units

    def _build_node_index(self):
        self._indexed = False
        for pop_name, pop_grp in self._population_map.items():
            sort_order = self._population_sorting_map[pop_name]
            nodes_indices = {}
            node_ids_ds = pop_grp[self._DATASET_node_ids]
            if sort_order == SortOrder.by_id:
                indx_beg = 0
                last_id = node_ids_ds[0]
                for indx, cur_id in enumerate(node_ids_ds):
                    if cur_id != last_id:
                        # nodes_indices[last_id] = np.arange(indx_beg, indx)
                        nodes_indices[last_id] = slice(indx_beg, indx)
                        last_id = cur_id
                        indx_beg = indx
                # nodes_indices[last_id] = np.arange(indx_beg, indx + 1)
                nodes_indices[last_id] = slice(indx_beg, indx + 1)  # capture the last node_id
            else:
                nodes_indices = {int(node_id): [] for node_id in np.unique(node_ids_ds)}
                for indx, node_id in enumerate(node_ids_ds):
                    nodes_indices[node_id].append(indx)

            self._index_nids[pop_name] = nodes_indices

        self._indexed = True

    @property
    def populations(self):
        return list(self._population_map.keys())

    def units(self, population=None):
        population = population if population is not None else self._default_pop
        return self._units_maps[population]

    def set_units(self, u, population=None):
        self._units_maps[population] = u

    def sort_order(self, population=None):
        return self._population_sorting_map[population]

    def node_ids(self, population=None):
        population = population if population is not None else self._default_pop
        pop_grp = self._population_map[population]
        return np.unique(pop_grp[self._DATASET_node_ids][()])

    def n_spikes(self, population=None):
        population = population if population is not None else self._default_pop
        return len(self._population_map[population][DATASET_timestamps])

    def time_range(self, populations=None):
        if populations is None:
            populations = [self._default_pop]

        if isinstance(populations, six.string_types) or np.isscalar(populations):
            populations = [populations]

        min_time = np.inf
        max_time = -np.inf
        for pop_name, pop_grp in self._population_map.items():
            if pop_name in populations:
                # TODO: Check if sorted by time
                for ts in pop_grp[DATASET_timestamps]:
                    if ts < min_time:
                        min_time = ts

                    if ts > max_time:
                        max_time = ts

        return min_time, max_time

    def to_dataframe(self, populations=None, sort_order=SortOrder.none, with_population_col=True, **kwargs):
        populations = populations if populations is not None else self.populations
        if isinstance(populations, six.string_types) or np.isscalar(populations):
            populations = [populations]

        ret_df = None
        for pop_name, pop_grp in self._population_map.items():
            if pop_name in populations:
                pop_df = pd.DataFrame({
                    col_timestamps: pop_grp[DATASET_timestamps],
                    # col_population: pop_name,
                    col_node_ids: pop_grp[self._DATASET_node_ids]
                })

                if with_population_col:
                    pop_df['population'] = pop_name

                if sort_order == SortOrder.by_id:
                    pop_df = pop_df.sort_values('node_ids')
                elif sort_order == SortOrder.by_time:
                    pop_df = pop_df.sort_values('timestamps')

                ret_df = pop_df if ret_df is None else pd.concat((ret_df, pop_df))

        if sort_order == SortOrder.by_time:
            ret_df.sort_values(by=col_timestamps, inplace=True)
        elif sort_order == SortOrder.by_id:
            ret_df.sort_values(by=col_node_ids, inplace=True)

        return ret_df

    def get_times(self, node_id, population=None, time_window=None, **kwargs):
        if population is None:
            if not isinstance(self._default_pop, six.string_types) and len(self._default_pop) > 1:
                raise Exception('Error: Multiple populations, must select one.')

            population = self._default_pop

        elif population not in self._population_map:
            return []

        spikes_index = self._index_nids[population].get(node_id, None)
        if spikes_index is None:
            return []
        spike_times = self._population_map[population][DATASET_timestamps][spikes_index]

        if time_window is not None:
            spike_times = spike_times[(time_window[0] <= spike_times) & (spike_times <= time_window[1])]

        return spike_times

    def spikes(self, node_ids=None, populations=None, time_window=None, sort_order=SortOrder.none, **kwargs):
        populations = populations or self.populations
        if np.isscalar(populations):
            populations = [populations]

        if sort_order == SortOrder.by_id:
            for pop_name in populations:
                if pop_name not in self.populations:
                    continue

                timestamps_ds = self._population_map[pop_name][DATASET_timestamps]
                index_map = self._index_nids[pop_name]
                node_ids = list(index_map.keys())
                node_ids.sort()
                for node_id in node_ids:
                    st_indices = index_map[node_id]
                    for st in timestamps_ds[st_indices]:  # st_indices:
                        yield st, pop_name, node_id

        elif sort_order == SortOrder.by_time:
            # TODO: Reimplement using a heap
            index_ranges = []
            for pop_name in populations:
                if pop_name not in self.populations:
                    continue

                pop_grp = self._population_map[pop_name]
                if self._population_sorting_map[pop_name] == SortOrder.by_time:
                    ts_ds = pop_grp[DATASET_timestamps]
                    index_ranges.append([pop_name, 0, len(ts_ds), np.arange(len(ts_ds)), pop_grp, ts_ds[0]])
                else:
                    ts_ds = pop_grp[DATASET_timestamps]
                    ts_indices = np.argsort(ts_ds[()])
                    index_ranges.append([pop_name, 0, len(ts_ds), ts_indices, pop_grp, ts_ds[ts_indices[0]]])

            while index_ranges:
                selected_r = index_ranges[0]
                for i, r in enumerate(index_ranges[1:]):
                    if r[5] < selected_r[5]:
                        selected_r = r

                ds_index = selected_r[1]
                timestamp = selected_r[5]  # pop_grp[DATASET_timestamps][ds_index]
                node_id = selected_r[4][self._DATASET_node_ids][ds_index]
                pop_name = selected_r[0]
                ds_index += 1
                if ds_index >= selected_r[2]:
                    index_ranges.remove(selected_r)
                else:
                    selected_r[1] = ds_index
                    ts_index = selected_r[3][ds_index]
                    next_ts = self._population_map[pop_name][DATASET_timestamps][selected_r[3][ds_index]]
                    selected_r[5] = next_ts  # pop_grp[DATASET_timestamps][selected_r[3][ds_index]]

                yield timestamp, pop_name, node_id

        else:
            for pop_name in populations:
                if pop_name not in self.populations:
                    continue

                pop_grp = self._population_map[pop_name]
                for i in range(len(pop_grp[DATASET_timestamps])):
                    yield pop_grp[DATASET_timestamps][i], pop_name, pop_grp[self._DATASET_node_ids][i]

    def __len__(self):
        if self._n_spikes is None:
            self._n_spikes = 0
            for _, pop_grp in self._population_map.items():
                self._n_spikes += len(pop_grp[self._DATASET_node_ids])

        return self._n_spikes


class SonataOldReader(SonataSTReader):
    """Older version of SONATA
    """

    def node_ids(self, population=None):
        return super(SonataOldReader, self).node_ids(population=None)

    def n_spikes(self, population=None):
        return super(SonataOldReader, self).n_spikes(population=self._default_pop)

    def time_range(self, populations=None):
        return super(SonataOldReader, self).time_range(populations=None)

    def to_dataframe(self, node_ids=None, populations=None, time_window=None, sort_order=SortOrder.none, **kwargs):
        return super(SonataOldReader, self).to_dataframe(node_ids=node_ids, populations=None,
                                                         time_window=time_window, sort_order=sort_order, **kwargs)

    def get_times(self, node_id, population=None, time_window=None, **kwargs):
        return super(SonataOldReader, self).get_times(node_id=node_id, population=None,
                                                      time_window=time_window, **kwargs)

    def spikes(self, node_ids=None, populations=None, time_window=None, sort_order=SortOrder.none, **kwargs):
        return super(SonataOldReader, self).spikes(node_ids=node_ids, populations=None,
                                                   time_window=time_window, sort_order=sort_order, **kwargs)


class EmptySonataReader(SpikeTrainsReadOnlyAPI):
    """A Hack that is needed for when a simulation produces a file with no spikes, since there won't/can't be
    <population_name> subgroup and/or gids/timestamps datasets.

    """
    def __init__(self, path, **kwargs):
        pass

    @property
    def populations(self):
        return []

    def node_ids(self, population=None):
        return []

    def n_spikes(self, population=None):
        return 0

    def time_range(self, populations=None):
        return None

    def to_dataframe(self, node_ids=None, populations=None, time_window=None, sort_order=SortOrder.none, **kwargs):
        return pd.DataFrame(columns=csv_headers)

    def get_times(self, node_id, population=None, time_window=None, **kwargs):
        return []

    def spikes(self, node_ids=None, populations=None, time_window=None, sort_order=SortOrder.none, **kwargs):
        return []


class CSVSTReader(SpikeTrainsReadOnlyAPI):
    def __init__(self, path, sep=' ', default_population=None, **kwargs):
        self._n_spikes = None
        self._populations = None

        try:
            # check to see if file contains headers
            with open(path, 'r') as csvfile:
                sniffer = csv.Sniffer()
                has_headers = sniffer.has_header(csvfile.read(1024))
        except Exception:
            has_headers = True

        self._spikes_df = pd.read_csv(path, sep=sep, header=0 if has_headers else None)

        if not has_headers:
            self._spikes_df.columns = csv_headers[0::2]

        self._defaul_population = default_population if default_population is not None \
            else self._spikes_df[col_population][0]

        if col_population not in self._spikes_df.columns:
            pop_name = kwargs.get(col_population, self._defaul_population)
            self._spikes_df[col_population] = pop_name

        # TODO: Check all the necessary columns exits
        self._spikes_df = self._spikes_df[csv_headers]

    @property
    def populations(self):
        if self._populations is None:
            self._populations = self._spikes_df['population'].unique()

        return self._populations

    def to_dataframe(self, populations=None, sort_order=SortOrder.none, with_population_col=True, **kwargs):
        selected = self._spikes_df.copy()

        mask = True
        if populations is not None:
            if isinstance(populations, six.string_types) or np.isscalar(populations):
                mask &= selected[col_population] == populations
            else:
                mask &= selected[col_population].isin(populations)

        if isinstance(mask, pd.Series):
            selected = selected[mask]

        if sort_order == SortOrder.by_time:
            selected.sort_values(by=col_timestamps, inplace=True)
        elif sort_order == SortOrder.by_id:
            selected.sort_values(by=col_node_ids, inplace=True)

        if not with_population_col:
            selected = selected.drop(col_population, axis=1)

        selected.index = pd.RangeIndex(len(selected.index))
        return selected

    def get_times(self, node_id, population=None, time_window=None, **kwargs):
        selected = self._spikes_df.copy()
        mask = (selected[col_node_ids] == node_id)

        if population is not None:
            mask &= (selected[col_population] == population)

        if time_window is not None:
            mask &= (selected[col_timestamps] >= time_window[0]) & (selected[col_timestamps] <= time_window[1])

        return np.array(self._spikes_df[mask][col_timestamps])

    def node_ids(self, population=None):
        population = population if population is not None else self._defaul_population
        return np.unique(self._spikes_df[self._spikes_df[col_population] == population][col_node_ids])

    def n_spikes(self, population=None):
        population = population if population is not None else self._defaul_population
        return len(self.to_dataframe(populations=population))

    # def time_range(self, populations=None):
    #     selected = self._spikes_df.copy()
    #     if populations is not None:
    #         if isinstance(populations, six.string_types) or np.isscalar(populations):
    #             mask = selected[col_population] == populations
    #         else:
    #             mask = selected[col_population].isin(populations)
    #
    #         selected = selected[mask]
    #
    #     return selected[col_timestamps].agg([np.min, np.max]).values

    def spikes(self, node_ids=None, populations=None, time_window=None, sort_order=SortOrder.none, **kwargs):
        selected = self._spikes_df.copy()

        mask = True
        if populations is not None:
            if isinstance(populations, six.string_types) or np.isscalar(populations):
                mask &= selected[col_population] == populations
            else:
                mask &= selected[col_population].isin(populations)

        if node_ids is not None:
            node_ids = [node_ids] if np.isscalar(node_ids) else node_ids
            mask &= selected[col_node_ids].isin(node_ids)

        if time_window is not None:
            mask &= (selected[col_timestamps] >= time_window[0]) & (selected[col_timestamps] <= time_window[1])

        if isinstance(mask, pd.Series):
            selected = selected[mask]

        if sort_order == SortOrder.by_time:
            selected.sort_values(by=col_timestamps, inplace=True)
        elif sort_order == SortOrder.by_id:
            selected.sort_values(by=col_node_ids, inplace=True)

        indicies = selected.index.values
        for indx in indicies:
            yield tuple(self._spikes_df.iloc[indx])

    def __len__(self):
        if self._n_spikes is None:
            self._n_spikes = len(self._spikes_df)

        return self._n_spikes


class NWBSTReader(SpikeTrainsReadOnlyAPI):
    def __init__(self, path, **kwargs):
        self._path = path
        self._h5_file = h5py.File(self._path, 'r')
        self._n_spikes = None
        self._spikes_df = None

        # TODO: Check for other versions
        self._population = kwargs.get('population', pop_na)
        if 'trial' in kwargs.keys():
            self._trial = kwargs['trial']
        elif len(self._h5_file['/processing']) == 1:
            self._trial = list(self._h5_file['/processing'].keys())[0]
        else:
            raise Exception('Please specify a trial')

        self._trial_grp = self._h5_file['processing'][self._trial]['spike_train']

    @property
    def populations(self):
        return [self._population]

    def node_ids(self, population=None):
        # if populations is None:
        #     populations = [self._population]
        # elif isinstance(populations, six.string_types):
        #     populations = [populations]

        if self._population != population:
            return []
        return [(self._population, np.uint64(node_id)) for node_id in self._trial_grp.keys()]

    def n_spikes(self, population=None):
        if population != self._population:
            return 0

        return self.__len__()

    # def time_range(self, populations=None):
    #     data_df = self.to_dataframe()
    #     return data_df[col_timestamps].agg([np.min, np.max]).values

    def get_times(self, node_id, population=None, time_window=None, **kwargs):
        try:
            spiketimes = self._trial_grp[str(node_id)]['data'][()]

            if time_window is not None:
                spiketimes = spiketimes[(time_window[0] <= spiketimes) & (spiketimes <= time_window[1])]

            return spiketimes
        except KeyError:
            return []

    def to_dataframe(self, node_ids=None, populations=None, time_window=None, sort_order=SortOrder.none, **kwargs):
        if self._spikes_df is None:
            self._spikes_df = pd.DataFrame({
                col_timestamps: pd.Series(dtype=np.float),
                col_population: pd.Series(dtype=np.string_),
                col_node_ids: pd.Series(dtype=np.uint64)
            })
            for node_id, node_grp in self._trial_grp.items():
                timestamps = node_grp['data'][()]
                node_df = pd.DataFrame({
                    col_timestamps: timestamps,
                    col_population: self._population,
                    col_node_ids: np.uint64(node_id)
                })
                self._spikes_df = self._spikes_df.append(node_df, ignore_index=True)

        selected = self._spikes_df.copy()
        mask = True
        if populations is not None:
            if isinstance(populations, six.string_types) or np.isscalar(populations):
                mask &= selected[col_population] == populations
            else:
                mask &= selected[col_population].isin(populations)

        if node_ids is not None:
            node_ids = [node_ids] if np.isscalar(node_ids) else node_ids
            mask &= selected[col_node_ids].isin(node_ids)

        if time_window is not None:
            mask &= (selected[col_timestamps] >= time_window[0]) & (selected[col_timestamps] <= time_window[1])

        if isinstance(mask, pd.Series):
            selected = selected[mask]

        if sort_order == SortOrder.by_time:
            selected.sort_values(by=col_timestamps, inplace=True)
        elif sort_order == SortOrder.by_id:
            selected.sort_values(by=col_node_ids, inplace=True)

        return selected

    def spikes(self, node_ids=None, populations=None, time_window=None, sort_order=SortOrder.none, **kwargs):
        if populations is not None:
            if np.isscalar(populations) and populations != self._population:
                raise StopIteration
            elif self._population not in populations:
                raise StopIteration

        if sort_order == SortOrder.by_time:
            spikes_df = self.to_dataframe()
            spikes_df = spikes_df.sort_values(col_timestamps)
            for indx in spikes_df.index:
                r = spikes_df.loc[indx]
                yield (r[col_timestamps], r[col_population], r[col_node_ids])
        else:
            node_ids = np.array(list(self._trial_grp.keys()), dtype=np.uint64)
            if sort_order == SortOrder.by_id:
                node_ids.sort()

            for node_id in node_ids:
                timestamps = self._trial_grp[str(node_id)]['data']
                for ts in timestamps:
                    yield (ts, self._population, node_id)

    def __len__(self):
        if self._n_spikes is None:
            self._n_spikes = 0
            for node_id in self._trial_grp.keys():
                self._n_spikes += len(self._trial_grp[node_id]['data'])

        return self._n_spikes



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import math

from bmtk.simulator.bionet.pyfunction_cache import add_weight_function


def default_weight_fnc(edge_props, src_props, trg_props):
    return edge_props['syn_weight']


def wmax(edge_props, src_props, trg_props):
    return edge_props["syn_weight"]


def gaussianLL(edge_props, src_props, trg_props):
    src_tuning = src_props['tuning_angle']
    tar_tuning = trg_props['tuning_angle']

    w0 = edge_props["syn_weight"]
    sigma = edge_props["weight_sigma"]

    delta_tuning = abs(abs(abs(180.0 - abs(float(tar_tuning) - float(src_tuning)) % 360.0) - 90.0) - 90.0)
    weight = w0 * math.exp(-(delta_tuning / sigma) ** 2)

    return weight


add_weight_function(wmax, 'wmax', overwrite=False)
add_weight_function(gaussianLL, 'gaussianLL', overwrite=False)
add_weight_function(default_weight_fnc, 'default_weight_fnc', overwrite=False)
add_weight_function(default_weight_fnc, 'set_syn_weight', overwrite=False)



import os
import re
import numpy as np
import scipy.io as sio
from scipy.fftpack import fft
import pandas as pd
from .movie import Movie, FullFieldFlashMovie


pd.set_option('display.width', 1000)
pd.set_option('display.max_columns', 100)


#################################################
def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]


##################################################
def compute_FFT_OneCycle(FR, TF, downsample):
    one_cyc = np.int(((1000. / downsample) / TF))
    FR_cyc = list(chunks(FR, one_cyc))
    if (TF == 15. or TF == 8.):
        FR_cyc = FR_cyc[:-1]

    FR_cyc_avg = np.mean(FR_cyc, axis=0)
    y = FR_cyc_avg
    AMP = 2 * np.abs(fft(y) / len(y))
    F0 = 0.5 * AMP[0]
    assert (F0 - np.mean(y) < 1.e-4)
    F1 = AMP[1]

    return F0, F1


##################################################
def create_ff_mov(frame_rate, tst, tend, xrng, yrng):
    ff_mov_on = FullFieldFlashMovie(np.arange(xrng), np.arange(yrng), tst, tend, frame_rate=frame_rate,
                                    max_intensity=1).full(t_max=tend)  # +0.5)
    ff_mov_off = FullFieldFlashMovie(np.arange(xrng), np.arange(yrng), tst, tend, frame_rate=frame_rate,
                                     max_intensity=-1).full(t_max=tend)  # +0.5)

    return ff_mov_on, ff_mov_off


##################################################
def create_grating_movie_list(gr_dir_name):
    gr_fnames = os.listdir(gr_dir_name)
    gr_fnames_ord = sorted(gr_fnames, key=lambda x: (int(re.sub('\D', '', x)), x))

    gr_mov_list = []
    for fname in gr_fnames_ord[:5]:
        movie_file = os.path.join(gr_dir_name, fname)
        m_file = sio.loadmat(movie_file)
        m_data_raw = m_file['mov'].T
        swid = np.shape(m_data_raw)[1]
        res = int(np.sqrt(swid / (8 * 16)))
        m_data = np.reshape(m_data_raw, (3000, 8 * res, 16 * res))
        m1 = Movie(m_data[:500, :, :], row_range=np.linspace(0, 120, m_data.shape[1], endpoint=True), col_range=np.linspace(0, 120, m_data.shape[2], endpoint=True), frame_rate=1000.)
        gr_mov_list.append(m1)

    return gr_mov_list

"""
##################################################
metrics_dir = os.path.join(os.path.dirname(__file__), 'cell_metrics')
def get_data_metrics_for_each_subclass(ctype):
    # Load csv file into dataframe
    if ctype.find('_sus') >= 0:
        prs_fn = os.path.join(metrics_dir, '{}_cells_v3.csv'.format(ctype))
    else:
        prs_fn = os.path.join(metrics_dir, '{}_cell_data.csv'.format(ctype))

    prs_df = pd.read_csv(prs_fn)
    N_class, nmet = np.shape(prs_df)

    # Group data by subclasses based on max F0 vals
    exp_df = prs_df.iloc[:, [13, 14, 17, 18, 28, 45, 46, 47, 48, 49, 50, 51, 52, 53,
                             54]].copy()  # Bl_lat,Wh_lat,Bl_si, wh_si, spont, 5 F0s, 5 F1s
    sub_df = exp_df.iloc[:, [5, 6, 7, 8, 9]]
    exp_df['max_tf'] = sub_df.idxmax(axis=1).values  # sub_df.idxmax(axis=1)

    exp_means = exp_df.groupby(['max_tf']).mean()
    exp_std = exp_df.groupby(['max_tf']).std()
    exp_nsub = exp_df.groupby(['max_tf']).size()

    max_ind_arr = np.where(exp_nsub == np.max(exp_nsub))
    max_nsub_ind = max_ind_arr[0][0]

    # Get means and std dev for subclasses
    exp_prs_dict = {}
    for scn in np.arange(len(exp_nsub)):
        f0_exp = exp_means.iloc[scn, 5:10].values
        f1_exp = exp_means.iloc[scn, 10:].values
        spont_exp = exp_means.iloc[scn, 4:5].values
        if ctype.find('OFF') >= 0:
            si_exp = exp_means.iloc[scn, 2:3].values
            ttp_exp = exp_means.iloc[scn, 0:1].values
        elif ctype.find('ON') >= 0:
            si_exp = exp_means.iloc[scn, 3:4].values
            ttp_exp = exp_means.iloc[scn, 1:2].values
        else:
            si_exp = np.NaN * np.ones((1, 5))
            ttp_exp = np.NaN * np.ones((1, 2))

        nsub = exp_nsub.iloc[scn]
        if nsub == 1:
            f0_std = np.mean(exp_std.iloc[max_nsub_ind, 5:10].values) * np.ones((1, 5))
            f1_std = np.mean(exp_std.iloc[max_nsub_ind, 10:].values) * np.ones((1, 5))
            spont_std = np.mean(exp_std.iloc[max_nsub_ind, 4:5].values) * np.ones((1, 5))
            if ctype.find('OFF') >= 0:
                si_std = np.mean(exp_std.iloc[max_nsub_ind, 2:3].values) * np.ones((1, 5))
            elif ctype.find('ON') >= 0:
                si_std = np.mean(exp_std.iloc[max_nsub_ind, 3:4].values) * np.ones((1, 5))
            else:
                si_std = np.NaN * np.ones((1, 5))

        else:
            f0_std = exp_std.iloc[scn, 5:10].values
            f1_std = exp_std.iloc[scn, 10:].values
            spont_std = exp_std.iloc[scn, 4:5].values
            if ctype.find('OFF') >= 0:
                si_std = exp_std.iloc[scn, 2:3].values
            elif ctype.find('ON') >= 0:
                si_std = exp_std.iloc[scn, 3:4].values
            else:
                si_std = np.NaN * np.ones((1, 5))

        if ctype.find('t') >= 0:
            tcross = 40.
            si_inf_exp = (si_exp - tcross / 200.) * (200. / (200. - tcross - 40.))
        elif ctype.find('s') >= 0:
            tcross = 60.
            si_inf_exp = (si_exp - tcross / 200.) * (200. / (200. - tcross - 40.))
        else:
            si_inf_exp = np.nan

        dict_key = exp_means.iloc[scn].name[3:]
        exp_prs_dict[dict_key] = {}
        exp_prs_dict[dict_key]['f0_exp'] = f0_exp
        exp_prs_dict[dict_key]['f1_exp'] = f1_exp
        exp_prs_dict[dict_key]['spont_exp'] = spont_exp
        exp_prs_dict[dict_key]['si_exp'] = si_exp
        exp_prs_dict[dict_key]['si_inf_exp'] = si_inf_exp
        exp_prs_dict[dict_key]['ttp_exp'] = ttp_exp
        exp_prs_dict[dict_key]['f0_std'] = f0_std
        exp_prs_dict[dict_key]['f1_std'] = f1_std
        exp_prs_dict[dict_key]['spont_std'] = spont_std
        exp_prs_dict[dict_key]['si_std'] = si_std
        exp_prs_dict[dict_key]['nsub'] = nsub
        exp_prs_dict[dict_key]['N_class'] = N_class

    return exp_prs_dict
"""


##################################################
def check_optim_results_against_bounds(bounds, opt_wts, opt_kpeaks):
    bds_wts0 = bounds[0]
    bds_wts1 = bounds[1]
    bds_kp0 = bounds[2]
    bds_kp1 = bounds[3]

    opt_wts0 = opt_wts[0]
    opt_wts1 = opt_wts[1]
    opt_kp0 = opt_kpeaks[0]
    opt_kp1 = opt_kpeaks[1]

    if (opt_wts0 == bds_wts0[0] or opt_wts0 == bds_wts0[1]):
        prm_on_bds = 'w0'
    elif (opt_wts1 == bds_wts1[0] or opt_wts1 == bds_wts1[1]):
        prm_on_bds = 'w1'
    elif (opt_kp0 == bds_kp0[0] or opt_kp0 == bds_kp0[1]):
        prm_on_bds = 'kp0'
    elif (opt_kp1 == bds_kp1[0] or opt_kp1 == bds_kp1[1]):
        prm_on_bds = 'kp1'
    else:
        prm_on_bds = 'None'

    return prm_on_bds


def cross_from_above(x, threshold):
    """Return the indices into *x* where *x* crosses some threshold from above."""
    x = np.asarray(x)
    ind = np.nonzero((x[:-1] >= threshold) & (x[1:] < threshold))[0]
    if len(ind):
        return ind+1
    else:
        return ind


#######################################################
def get_tcross_from_temporal_kernel(temporal_kernel):
    max_ind = np.argmax(temporal_kernel)
    min_ind = np.argmin(temporal_kernel)

    temp_tcross_ind = cross_from_above(temporal_kernel[max_ind:min_ind], 0.0)
    tcross_ind = max_ind + temp_tcross_ind[0]
    return tcross_ind



import numpy as np
from sympy.abc import x as symbolic_x
from sympy.abc import y as symbolic_y

from bmtk.simulator.filternet.filters import TemporalFilterCosineBump, GaussianSpatialFilter, SpatioTemporalFilter
from bmtk.simulator.filternet.cell_models import TwoSubfieldLinearCell, OnUnit, OffUnit, LGNOnOffCell
from bmtk.simulator.filternet.transfer_functions import ScalarTransferFunction, MultiTransferFunction
from bmtk.simulator.filternet.utils import get_data_metrics_for_each_subclass, get_tcross_from_temporal_kernel
from bmtk.simulator.filternet.pyfunction_cache import py_modules


def create_two_sub_cell(dom_lf, non_dom_lf, dom_spont, non_dom_spont, onoff_axis_angle, subfield_separation,
                        dom_location):
    dsp = str(dom_spont)
    ndsp = str(non_dom_spont)
    two_sub_transfer_fn = MultiTransferFunction((symbolic_x, symbolic_y),
                                                'Heaviside(x+'+dsp+')*(x+'+dsp+')+Heaviside(y+'+ndsp+')*(y+'+ndsp+')')

    two_sub_cell = TwoSubfieldLinearCell(dom_lf, non_dom_lf, subfield_separation=subfield_separation,
                                         onoff_axis_angle=onoff_axis_angle, dominant_subfield_location=dom_location,
                                         transfer_function=two_sub_transfer_fn)
    return two_sub_cell


def createOneUnitOfTwoSubunitFilter(weights, kpeaks, delays, ttp_exp):
    delays = np.array(delays)
    filt = TemporalFilterCosineBump(weights, kpeaks, delays)
    tcross_ind = get_tcross_from_temporal_kernel(filt.get_kernel(threshold=-1.0).kernel)
    filt_sum = filt.get_kernel(threshold=-1.0).kernel[:tcross_ind].sum()

    # Calculate delay offset needed to match response latency with data and rebuild temporal filter
    del_offset = ttp_exp - tcross_ind
    if del_offset >= 0:
        delays_updated = delays + del_offset
        filt_new = TemporalFilterCosineBump(weights, kpeaks, delays_updated)
    else:
        raise Exception('del_offset < 0')

    return filt_new, filt_sum


def get_tf_params(node, dynamics_params, non_dom_props=False):
    if not non_dom_props:
        weights = node.weights if node.weights is not None else dynamics_params['opt_wts']
        kpeaks = node.kpeaks if node.kpeaks is not None else dynamics_params['opt_kpeaks']
        delays = node.delays if node.delays is not None else dynamics_params['opt_delays']
    else:
        dp = dynamics_params or {}
        weights = node.weights_non_dom if node.weights_non_dom is not None else dp.get('opt_wts', None)
        kpeaks = node.kpeaks_non_dom if node.kpeaks_non_dom is not None else dp.get('opt_kpeaks', None)
        delays = node.delays_non_dom if node.delays_non_dom is not None else dp.get('opt_delays', None)

    if node.predefined_jitter:
        jitter_fnc = lambda a: np.array([np.random.uniform(x*node.jitter[0], x*node.jitter[1]) for x in a])
        weights = jitter_fnc(weights) if weights is not None else weights
        kpeaks = jitter_fnc(kpeaks) if kpeaks is not None else kpeaks
        delays = jitter_fnc(delays) if delays is not None else delays

    return weights, kpeaks, delays


def get_sigma(node, dynamics_params):
    if 'spatial_size' in node:
        sigma = node['spatial_size']
    elif 'sigma' in node:
        sigma = node['sigma']
    elif 'spatial_size' in dynamics_params:
        sigma = dynamics_params['spatial_size']
    elif 'sigma' in dynamics_params:
        sigma = dynamics_params['spatial_size']
    else:
        # TODO: Raise warning
        sigma = (1.0, 1.0)

    if np.isscalar(sigma):
        sigma = (sigma, sigma) # convert from degree to SD

    return sigma[0]/3.0, sigma[1]/3.0


def default_cell_loader(node, template_name, dynamics_params):
    """

    :param node:
    :param template_name:
    :param dynamics_params:
    :return:
    """
    # Create the spatial filter
    origin = (0.0, 0.0)
    translate = (node['x'], node['y'])

    sigma = get_sigma(node, dynamics_params)
    if 'spatial_rotation' in node:
        rotation = node['spatial_rotation']
    else:
        rotation = 0.0

    spatial_filter = GaussianSpatialFilter(translate=translate, sigma=sigma, origin=origin, rotation=rotation)

    t_weights, t_kpeaks, t_delays = get_tf_params(node, dynamics_params)

    if template_name:
        model_name = template_name[1]
    else:
        model_name = node['pop_name']

    if model_name in ['sONsOFF_001', 'sONsOFF']:
        # sON temporal filter
        t_weights_nd, t_kpeaks_nd, t_delays_nd = get_tf_params(node, node.non_dom_params, non_dom_props=True)
        sON_filt_new, sON_sum = createOneUnitOfTwoSubunitFilter(t_weights_nd, t_kpeaks_nd, t_delays_nd, 121.0)
        sOFF_filt_new, sOFF_sum = createOneUnitOfTwoSubunitFilter(t_weights, t_kpeaks, t_delays, 115.0)

        amp_on = 1.0  # set the non-dominant subunit amplitude to unity
        spont = 4.0
        max_roff = 35.0
        max_ron = 21.0
        amp_off = -(max_roff/max_ron)*(sON_sum/sOFF_sum)*amp_on - (spont*(max_roff - max_ron))/(max_ron*sOFF_sum)

        # Create sON subunit:
        linear_filter_son = SpatioTemporalFilter(spatial_filter, sON_filt_new, amplitude=amp_on)

        # Create sOFF subunit:
        linear_filter_soff = SpatioTemporalFilter(spatial_filter, sOFF_filt_new, amplitude=amp_off)

        sf_sep = node.sf_sep
        if node.predefined_jitter:
            sf_sep = np.random.uniform(node.jitter[0]*sf_sep, node.jitter[1]*sf_sep)

        sep_ss_onoff_cell = create_two_sub_cell(linear_filter_soff, linear_filter_son, 0.5 * spont, 0.5 * spont,
                                                node.tuning_angle, sf_sep, translate)
        cell = sep_ss_onoff_cell

    elif model_name in ['sONtOFF_001', 'sONtOFF']:
        t_weights_nd, t_kpeaks_nd, t_delays_nd = get_tf_params(node, node.non_dom_params, non_dom_props=True)
        sON_filt_new, sON_sum = createOneUnitOfTwoSubunitFilter(t_weights_nd, t_kpeaks_nd, t_delays_nd, 93.5)

        tOFF_filt_new, tOFF_sum = createOneUnitOfTwoSubunitFilter(t_weights, t_kpeaks, t_delays, 64.8)  # 64.8

        amp_on = 1.0  # set the non-dominant subunit amplitude to unity
        spont = 5.5
        max_roff = 46.0
        max_ron = 31.0
        amp_off = -0.7*(max_roff/max_ron)*(sON_sum/tOFF_sum)*amp_on - (spont*(max_roff - max_ron))/(max_ron*tOFF_sum)

        # Create sON subunit:
        linear_filter_son = SpatioTemporalFilter(spatial_filter, sON_filt_new, amplitude=amp_on)

        # Create tOFF subunit:
        linear_filter_toff = SpatioTemporalFilter(spatial_filter, tOFF_filt_new, amplitude=amp_off)

        sf_sep = node.sf_sep
        if node.predefined_jitter:
            sf_sep = np.random.uniform(node.jitter[0]*sf_sep, node.jitter[1]*sf_sep)

        sep_ts_onoff_cell = create_two_sub_cell(linear_filter_toff, linear_filter_son, 0.5 * spont, 0.5 * spont,
                                                node.tuning_angle, sf_sep, translate)

        cell = sep_ts_onoff_cell

    elif model_name == 'LGNOnOFFCell':
        wts = [node['weight_dom_0'], node['weight_dom_1']]
        kpeaks = [node['kpeaks_dom_0'], node['kpeaks_dom_1']]
        delays = [node['delay_dom_0'], node['delay_dom_1']]
        # transfer_function = ScalarTransferFunction('s')
        temporal_filter = TemporalFilterCosineBump(wts, kpeaks, delays)

        spatial_filter_on = GaussianSpatialFilter(sigma=node['sigma_on'], origin=origin, translate=translate)
        on_linear_filter = SpatioTemporalFilter(spatial_filter_on, temporal_filter, amplitude=20)

        spatial_filter_off = GaussianSpatialFilter(sigma=node['sigma_off'], origin=origin, translate=translate)
        off_linear_filter = SpatioTemporalFilter(spatial_filter_off, temporal_filter, amplitude=-20)
        cell = LGNOnOffCell(on_linear_filter, off_linear_filter)

    else:
        type_split = model_name.split('_')

        if len(type_split) == 1:
            cell_type = model_name
            tf_str = 'TF8'
        else:
            cell_type, tf_str = type_split[0], type_split[1]

        # Get spontaneous firing rate, either from the cell property of calculate from experimental data
        if 'spont_fr' in node:
            spont_fr = node['spont_fr']
        else:
            exp_prs_dict = get_data_metrics_for_each_subclass(cell_type)
            subclass_prs_dict = exp_prs_dict[tf_str]
            spont_fr = subclass_prs_dict['spont_exp'][0]

        # Get filters
        transfer_function = ScalarTransferFunction('Heaviside(s+{})*(s+{})'.format(spont_fr, spont_fr))
        temporal_filter = TemporalFilterCosineBump(t_weights, t_kpeaks, t_delays)

        if cell_type.find('ON') >= 0:
            amplitude = 1.0
            linear_filter = SpatioTemporalFilter(spatial_filter, temporal_filter, amplitude=amplitude)
            cell = OnUnit(linear_filter, transfer_function)
        elif cell_type.find('OFF') >= 0:
            amplitude = -1.0
            linear_filter = SpatioTemporalFilter(spatial_filter, temporal_filter, amplitude=amplitude)
            cell = OffUnit(linear_filter, transfer_function)

    return cell


py_modules.add_cell_processor('default', default_cell_loader, overwrite=False)
py_modules.add_cell_processor('preset_params', default_cell_loader, overwrite=False)



from .memory_sorter import quicksort_edges
from .merge_sorter import external_merge_sort


def sort_edges(input_edges_path, output_edges_path, edges_population, sort_by, sort_model_properties=True,
               sort_on_disk=False, **sorter_args):
    if not sort_on_disk:
        quicksort_edges(
            input_edges_path=input_edges_path,
            output_edges_path=output_edges_path,
            edges_population=edges_population,
            sort_by=sort_by,
            sort_model_properties=sort_model_properties,
            **sorter_args
        )
    else:
        external_merge_sort(
            input_edges_path=input_edges_path,
            output_edges_path=output_edges_path,
            edges_population=edges_population,
            sort_by=sort_by,
            sort_model_properties=sort_model_properties,
            **sorter_args
        )



import numpy as np
from collections import Counter
import numbers
import nest
import types
import pandas as pd

from bmtk.simulator.core.sonata_reader import NodeAdaptor, SonataBaseNode, EdgeAdaptor, SonataBaseEdge
from bmtk.simulator.pointnet.io_tools import io
from bmtk.simulator.pointnet.pyfunction_cache import py_modules
from bmtk.simulator.pointnet.glif_utils import convert_aibs2nest
from bmtk.simulator.pointnet.nest_utils import nest_version

NEST_SYNAPSE_MODEL_PROP = 'model' if nest_version[0] == 2 else 'synapse_model'



def all_null(node_group, column_name):
    """Helper function to determine if a column has any non-NULL values"""
    types_table = node_group.parent.types_table
    non_null_vals = [types_table[ntid][column_name] for ntid in np.unique(node_group.node_type_ids)
                     if types_table[ntid][column_name] is not None]
    return len(non_null_vals) == 0


class PointNodeBatched(object):
    def __init__(self, node_ids, gids, node_types_table, node_type_id):
        self._n_nodes = len(node_ids)
        self._node_ids = node_ids
        self._gids = gids
        self._nt_table = node_types_table
        self._nt_id = node_type_id
        self._nest_objs = []
        self._nest_ids = []

    @property
    def n_nodes(self):
        return self._n_nodes

    @property
    def node_ids(self):
        return self._node_ids

    @property
    def gids(self):
        return self._gids

    @property
    def nest_ids(self):
        return self._nest_ids

    @property
    def nest_model(self):
        return self._nt_table[self._nt_id]['model_template'].split(':')[1]

    @property
    def nest_params(self):
        return self._nt_table[self._nt_id]['dynamics_params']

    @property
    def model_type(self):
        return self._nt_table[self._nt_id]['model_type']

    def build(self):
        self._nest_objs = nest.Create(self.nest_model, self.n_nodes, self.nest_params)
        self._nest_ids = self._nest_objs.tolist() if nest_version[0] >= 3 else self._nest_objs


class PointNode(SonataBaseNode):
    def __init__(self, node, prop_adaptor):
        super(PointNode, self).__init__(node, prop_adaptor)
        self._nest_objs = []
        self._nest_ids = []

    @property
    def n_nodes(self):
        return 1

    @property
    def node_ids(self):
        return [self._prop_adaptor.node_id(self._node)]

    @property
    def gids(self):
        return [self._prop_adaptor.gid(self._node)]

    @property
    def nest_ids(self):
        return self._nest_ids

    @property
    def nest_model(self):
        return self._prop_adaptor.model_template(self._node)[1]

    @property
    def nest_params(self):
        return self.dynamics_params

    def build(self):
        nest_model = self.nest_model
        dynamics_params = self.dynamics_params
        fnc_name = self._node['model_processing']
        if fnc_name is None:
            self._nest_objs = nest.Create(nest_model, 1, dynamics_params)
        else:
            cell_fnc = py_modules.cell_processor(fnc_name)
            self._nest_objs = cell_fnc(nest_model, self._node, dynamics_params)

        self._nest_ids = self._nest_objs.tolist() if nest_version[0] >= 3 else self._nest_objs


class PointNodeAdaptor(NodeAdaptor):
    def __init__(self, network):
        super(PointNodeAdaptor, self).__init__(network)

        # Flag for determining if we can build multiple NEST nodes at once. If each individual node has unique
        # NEST params or a model_processing function is being called then we must nest.Create for each individual cell.
        # Otherwise we can try to call nest.Create for a batch of nodes that share the same properties
        self._can_batch = True

    @property
    def batch_process(self):
        return self._can_batch

    @batch_process.setter
    def batch_process(self, flag):
        self._can_batch = flag

    def get_node(self, sonata_node):
        return PointNode(sonata_node, self)

    def get_batches(self, node_group):
        node_ids = node_group.node_ids
        node_type_ids = node_group.node_type_ids
        node_gids = node_group.gids
        if node_gids is None:
            node_gids = node_ids

        ntids_counter = Counter(node_type_ids)

        nid_groups = {nt_id: np.zeros(ntids_counter[nt_id], dtype=np.uint32) for nt_id in ntids_counter}
        gid_groups = {nt_id: np.zeros(ntids_counter[nt_id], dtype=np.uint32) for nt_id in ntids_counter}
        node_groups_counter = {nt_id: 0 for nt_id in ntids_counter}

        for node_id, gid, node_type_id in zip(node_ids, node_gids, node_type_ids):
            grp_indx = node_groups_counter[node_type_id]
            nid_groups[node_type_id][grp_indx] = node_id
            gid_groups[node_type_id][grp_indx] = gid
            node_groups_counter[node_type_id] += 1

        return [PointNodeBatched(nid_groups[nt_id], gid_groups[nt_id], node_group.parent.node_types_table, nt_id)
                for nt_id in ntids_counter]

    @staticmethod
    def preprocess_node_types(network, node_population):
        NodeAdaptor.preprocess_node_types(network, node_population)
        node_types_table = node_population.types_table
        if 'model_template' in node_types_table.columns and 'dynamics_params' in node_types_table.columns:
            node_type_ids = np.unique(node_population.type_ids)
            for nt_id in node_type_ids:
                node_type_attrs = node_types_table[nt_id]
                mtemplate = node_type_attrs['model_template']
                dyn_params = node_type_attrs['dynamics_params']
                if mtemplate.startswith('nest:glif') and dyn_params.get('type', None) == 'GLIF':
                    model_template, dynamics_params = convert_aibs2nest(mtemplate, dyn_params)
                    node_type_attrs['model_template'] = model_template
                    node_type_attrs['dynamics_params'] = dynamics_params

    @staticmethod
    def patch_adaptor(adaptor, node_group, network):
        node_adaptor = NodeAdaptor.patch_adaptor(adaptor, node_group, network)

        # If dynamics params is stored in the nodes.h5 then we have to build each node separate
        if node_group.has_dynamics_params:
            node_adaptor.batch_process = False

        # If there is a non-null value in the model_processing column then it potentially means that every cell is
        # uniquly built (currently model_processing is applied to each individ. cell) and nodes can't be batched
        if 'model_processing' in node_group.columns:
            node_adaptor.batch_process = False
        elif 'model_processing' in node_group.all_columns and not all_null(node_group, 'model_processing'):
            node_adaptor.batch_process = False

        if node_adaptor.batch_process:
            io.log_info('Batch processing nodes for {}/{}.'.format(node_group.parent.name, node_group.group_id))

        return node_adaptor


class PointEdge(SonataBaseEdge):
    @property
    def source_node_ids(self):
        return [self._edge.source_node_id]

    @property
    def target_node_ids(self):
        return [self._edge.target_node_id]

    @property
    def nest_params(self):
        if self.model_template in py_modules.synapse_models:
            src_node = self._prop_adaptor._network.get_node_id(self.source_population, self.source_node_id)
            trg_node = self._prop_adaptor._network.get_node_id(self.target_population, self.target_node_id)
            syn_model_fnc = py_modules.synapse_model(self.model_template)
        else:
            src_node = None
            trg_node = None
            syn_model_fnc = py_modules.synapse_models('default')

        return syn_model_fnc(self, src_node, trg_node)


class PointEdgeBatched(object):
    def __init__(self, source_nids, target_nids, nest_params):
        self._src_nids = source_nids
        self._trg_nids = target_nids
        self._nest_params = nest_params

    @property
    def source_node_ids(self):
        return self._src_nids

    @property
    def target_node_ids(self):
        return self._trg_nids

    @property
    def nest_params(self):
        return self._nest_params


class PointEdgeAdaptor(EdgeAdaptor):
    def __init__(self, network):
        super(PointEdgeAdaptor, self).__init__(network)
        self._can_batch = True

    @property
    def batch_process(self):
        return self._can_batch

    @batch_process.setter
    def batch_process(self, flag):
        self._can_batch = flag

    def synaptic_params(self, edge):
        # TODO: THIS NEEDS to be replaced with call to synapse_models
        params_dict = {'weight': self.syn_weight(edge, None, None), 'delay': edge.delay}
        params_dict.update(edge.dynamics_params)
        return params_dict

    def get_edge(self, sonata_node):
        return PointEdge(sonata_node, self)

    @staticmethod
    def preprocess_edge_types(network, edge_population):
        # Fix for sonata/300_pointneurons
        EdgeAdaptor.preprocess_edge_types(network, edge_population)
        edge_types_table = edge_population.types_table
        edge_type_ids = np.unique(edge_population.type_ids)

        for et_id in edge_type_ids:
            edge_type = edge_types_table[et_id]
            if 'model_template' in edge_types_table.columns:
                model_template = edge_type['model_template']
                if model_template.startswith('nest'):
                    edge_type['model_template'] = model_template[5:]

    def get_batches(self, edge_group):
        src_ids = {}
        trg_ids = {}

        edge_types_table = edge_group.parent.edge_types_table

        edge_type_ids = edge_group.edge_type_ids
        et_id_counter = Counter(edge_type_ids)
        tmp_df = pd.DataFrame({'etid': edge_type_ids, 'src_nids': edge_group.src_node_ids,
                               'trg_nids': edge_group.trg_node_ids})

        if 'nsyns' in edge_group.columns:
            tmp_df['nsyns'] = edge_group.get_dataset('nsyns')

        if 'syn_weight' in edge_group.columns:
            tmp_df['syn_weight'] = edge_group.get_dataset('syn_weight')

        if 'delay' in edge_group.columns:
            tmp_df['delay'] = edge_group.get_dataset('delay')

        #for et_id, grp_vals in tmp_df.groupby('etid'):
        #    src_ids[et_id] = np.array(grp_vals['src_nids'])
        #    trg_ids[et_id] = np.array(grp_vals['trg_nids'])

        type_params = {edge_id: {} for edge_id in et_id_counter.keys()}
        src_pop = edge_group.parent.source_population
        trg_pop = edge_group.parent.target_population

        grp_df = None
        src_nodes_df = None
        trg_nodes_df = None
        for edge_id, grp_vals in tmp_df.groupby('etid'):
            edge_props = edge_types_table[edge_id]
            n_edges = len(grp_vals)

            # Get the model type
            type_params[edge_id][NEST_SYNAPSE_MODEL_PROP] = edge_props['model_template']

            # Add dynamics params
            # TODO: Add to dataframe and if a part of hdf5 we can return any dynamics params as a list
            type_params[edge_id].update(edge_props['dynamics_params'])

            # get the delay parameter
            if 'delay' in grp_vals.columns:
                type_params[edge_id]['delay'] = grp_vals['delay']
            elif 'delay' in edge_props.keys():
                delay = edge_props['delay']
                # For NEST 2.* 'delay' can be a single value, but for 3.* it requires a full array for each edge
                delay = np.full(n_edges, delay) if nest_version[0] >= 3 else delay
                type_params[edge_id]['delay'] = delay

            weight_function = edge_types_table[edge_id].get('weight_function', None)
            if weight_function is not None:
                if grp_df is None:
                    grp_df = edge_group.to_dataframe()
                    src_nodes_df = self._network.get_nodes_df(src_pop)
                    trg_nodes_df = self._network.get_nodes_df(trg_pop)
                edges = grp_df[grp_df['edge_type_id'] == edge_id]
                target_nodes = trg_nodes_df.loc[edges['target_node_id'].values]
                source_nodes = src_nodes_df.loc[edges['source_node_id'].values]

                if not py_modules.has_synaptic_weight(weight_function):
                    err_msg = 'Unable to calculate synaptic weight for "{}" edges, missing "weight_function" ' \
                              'attribute value {} function.'.format(edge_group.parent.name, weight_function)
                    io.log_exception(err_msg)

                weight_fnc = py_modules.synaptic_weight(weight_function)
                type_params[edge_id]['weight'] = weight_fnc(edges, source_nodes, target_nodes)

            else:
                # Get nsyns as either an array or a constant. If not explcitiy specified assume nsyns = 1
                if 'nsyns' in grp_vals.columns:
                    nsyns = grp_vals['nsyns'].values
                else:
                    nsyns = edge_props.get('nsyns', 1)

                # get syn_weight as either an array or constant. If not explicity stated throw an error
                if 'syn_weight' in grp_vals.columns:
                    syn_weight = grp_vals['syn_weight'].values
                elif 'syn_weight' in edge_props.keys():
                    syn_weight = edge_props['syn_weight']
                else:
                    # TODO: Make more explicity. Or default to syn_weight of 0
                    raise Exception('Could not find syn_weight value')

                # caluclate weight
                type_params[edge_id]['weight'] = nsyns * syn_weight

            yield PointEdgeBatched(source_nids=grp_vals['src_nids'].values, target_nids=grp_vals['trg_nids'].values,
                                   nest_params=type_params[edge_id])

    @staticmethod
    def patch_adaptor(adaptor, edge_group):
        edge_adaptor = EdgeAdaptor.patch_adaptor(adaptor, edge_group)

        if 'weight_function' not in edge_group.all_columns and 'syn_weight' in edge_group.all_columns:
            adaptor.syn_weight = types.MethodType(point_syn_weight, adaptor)

        #else:
        #    edge_adaptor.batch_process = False

        return edge_adaptor


def point_syn_weight(self, edge, src_node, trg_node):
    return edge['syn_weight']*edge.nsyns



import os
from six import string_types
import numpy as np

from bmtk.utils import sonata
from bmtk.utils.sonata.config import SonataConfig
from bmtk.utils.reports.compartment import CompartmentReport
from bmtk.utils.reports.compartment import plotting
from bmtk.simulator.utils import simulation_reports


def _get_report(report_path=None, config=None, report_name=None):
    if report_path is not None:
        return report_path, CompartmentReport.load(report=report_path)

    elif config is not None:
        selected_reports = []
        sim_reports = simulation_reports.from_config(config)
        for report in sim_reports:
            if report.module in ['membrane_report', 'multimeter_report']:
                rname = report.report_name
                rfile = report.params['file_name']
                # TODO: Full path should be determined by config/simulation_reports module
                rpath = rfile if os.path.isabs(rfile) else os.path.join(report.params['tmp_dir'], rfile)
                if report_name is not None and report_name == rname:
                    selected_reports.append((rname, CompartmentReport.load(rpath)))
                elif report_name is None:
                    selected_reports.append((rname, CompartmentReport.load(rpath)))

        if len(selected_reports) == 0:
            msg = 'Could not find a report '
            msg += '' if report_name is None else 'with report_name "{}"'.format(report_name)
            msg += ' from configuration file. . Use "report_path" parameter instead.'
            raise ValueError(msg)

        elif len(selected_reports) > 1:
            avail_reports = ', '.join(s[0] for s in selected_reports)
            raise ValueError('Configuration file contained multiple "membrane_reports", use "report_name" or'
                             '"report_path" to pick which one to plot. Option values: {}'.format(avail_reports))

        else:
            return selected_reports[0]

    else:
        raise AttributeError('Could not find a compartment report SONATA file. Please user "config_file" or '
                             '"report_path" options.')


def _find_nodes(population, config=None, nodes_file=None, node_types_file=None):
    if nodes_file is not None:
        network = sonata.File(data_files=nodes_file, data_type_files=node_types_file)
        if population not in network.nodes.population_names:
            raise ValueError('node population "{}" not found in {}'.format(population, nodes_file))
        return network.nodes[population]

    elif config is not None:
        for nodes_grp in config.nodes:
            network = sonata.File(data_files=nodes_grp['nodes_file'], data_type_files=nodes_grp['node_types_file'])
            if population in network.nodes.population_names:
                return network.nodes[population]

    raise ValueError('Could not find nodes file with node population "{}".'.format(population))


def plot_traces(config_file=None, report_name=None, population=None, report_path=None, group_by=None,
                group_excludes=None, nodes_file=None, node_types_file=None,
                node_ids=None, sections='origin', average=False, times=None, title=None,
                show_legend=None, show=True, save_as=None, plt_style=None):
    """Plot compartment variables (eg Membrane Voltage, Calcium conc.) traces from the output of simulation. Will
    attempt to look in the SONATA simulation configuration json "reports" sections for any matching "membrane_report"
    outputs with a matching report_name::

        plot_traces(config_file='config.json', report_name='membrane_potential')

    If the path the the report is different (or missing) than what's in the SONATA config then use the "report_path"
    option instead::

        plot_traces(report_path='/my/path/to/membrane_potential.h5')

    To display the traces of only a select number of nodes you can filter using the node_ids options::

        plot_traces(config_file='config.json', node_ids=[10, 20, 30, 40, 50])

    The average option will find the mean value of all the traces to display::

        plot_traces(config_file='config.json', node_ids=range(50, 100), average=True)

    You may also group together different subsets of nodes and display multiple averages based on certain attributes
    of the network, which can be done using the group_by key. The group_exlcudes option will exclude certain groups.
    For example if you want to plot the averaged membrane potential across each cortical "layer", exclude L1::

        plot_traces(config_file='config.json', report='membrane_potential', group_by='layer', group_excludes='L1')

    :param config_file: path to SONATA simulation configuration.
    :param report_name: name of the membrane_report "report" which will be plotted. If only one compartment report
        in the simulation config then function will find it automatically.
    :param population: string. If the report more than one population of nodes, use this to determine which nodes to
           plot. If only one population exists and population=None then the function will find it by default.
    :param report_path: Path to SONATA compartment report file. Do not use with "config_file" and "report_name" options.
    :param group_by: Attribute of the "nodes" file used to group and average subsets of nodes.
    :param group_excludes: list of strings or None. When using the "group_by", allows users to exclude certain groupings
        based on the attribute value.
    :param nodes_file: path to nodes hdf5 file containing "population". By default this will be resolved using the
        config.
    :param node_types_file: path to node-types csv file containing "population". By default this will be resolved using
        the config.
    :param node_ids: int or list of integers. Individual node to display the variable.
    :param sections: 'origin', 'all', or list of ids, Compartments/elements to display, By default will only show values
           at the soma.
    :param average: If true will display average of "node_ids". Default: False
    :param times: (float, float), start and stop times of simulation. By default will get values from simulation
        configs "run" section.
    :param title: str, adds a title to the plot. If None (default) then name will be automatically generated using the
        report_name.
    :param show_legend: Set True or False to determine if legend should be displayed on the plot. The default (None)
           function itself will guess if legend should be shown.
    :param show: bool to display or not display plot. default True.
    :param save_as: None or str: file-name/path to save the plot as a png/jpeg/etc. If None or empty string will not
        save plot.
    :return: matplotlib figure.Figure object
    """

    sonata_config = SonataConfig.from_json(config_file) if config_file else None
    report_name, cr = _get_report(report_path=report_path, config=sonata_config, report_name=report_name)

    if population is None:
        pops = cr.populations
        if len(pops) > 1:
            raise ValueError('Report {} contains more than population of nodes ({}). Use population parameter'.format(
                report_name, pops
            ))
        population = pops[0]

    if title is None:
        title = '{} ({})'.format(report_name, population)

    # Create node-groups
    if group_by is not None:
        node_groups = []
        nodes = _find_nodes(population=population, config=sonata_config, nodes_file=nodes_file,
                            node_types_file=node_types_file)

        grouped_df = None
        for grp in nodes.groups:
            if group_by in grp.all_columns:
                grp_df = grp.to_dataframe()
                grp_df = grp_df[['node_id', group_by]]
                grouped_df = grp_df if grouped_df is None else grouped_df.append(grp_df, ignore_index=True)

        if grouped_df is None:
            raise ValueError('Could not find any nodes with group_by attribute "{}"'.format(group_by))

        # Convert from string to list so we can always use the isin() method for filtering
        if isinstance(group_excludes, string_types):
            group_excludes = [group_excludes]
        elif group_excludes is None:
            group_excludes = []

        for grp_key, grp in grouped_df.groupby(group_by):
            if grp_key in group_excludes:
                continue
            node_groups.append({'node_ids': np.array(grp['node_id']), 'label': grp_key})

        if len(node_groups) == 0:
            exclude_str = ' excluding values {}'.format(', '.join(group_excludes)) if len(group_excludes) > 0 else ''
            raise ValueError('Could not find any node-groups using group_by="{}"{}.'.format(group_by, exclude_str))

    else:
        node_groups = None

    return plotting.plot_traces(
        report=cr,
        population=population,
        node_ids=node_ids,
        sections=sections,
        average=average,
        node_groups=node_groups,
        times=times,
        title=title,
        show_legend=show_legend,
        show=show,
        save_as=save_as,
        plt_style=plt_style
    )



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
from . import connector
from . import iterator


class ConnectionMap(object):
    """Class for keeping track of connection rules. Used for saving unique edge properties in a network

    For every connection from source --> target this keeps track of rules (functions, literals, lists) for
      1. the number of synapses between source and target
      2. Used defined parameters (syn-weight, synaptic-location) for every synapse.

    The number of synapses rule (1) is stored as a connector. Individual synaptic parameters, if they exists, are stored
    as ParamsRules.

    Usage:
        Do not create a ConnectionMap object directly. Rather one is created and return every time NetworkBuilder's
        add_edges() method is called.

        For example, The following will create a set of edges where there is between 1 to 16 synapatic connections
        between each source and target node. Properties 'dynamics_params' and 'model_template' will be global
        properties::

            cm = net.add_edges(source={'ei': 'exc'}, target={'ei': 'inh'},
                               connection_rule=lambda *_: np.random.randn(1, 16),
                               dynamics_params='ExcToInh.json',
                               model_template='Exp2Syn',
                               ...)

        If we want all edges to have a unique 'syn_weight' value calculated from source/target locations, we create a
        function that takes in source and target (Node object) and returns the syn weight. This function will be called
        for every individual connection::

            def syn_weight_by_loc(source, target):
                src_loc, trg_loc = source['loc'], target['loc']
                ...
                return syn_weight

            cm.add_properties('syn_weight', rule=syn_weight_by_loc, dtypes=np.float)

        We can also add multiple properties at the same time::

            def get_target_loc_and_dist(source, target):
                ...
                return trg_loc, trg_dist

            cm.add_properties(['syn_location', 'syn_distance'],
                          rule=get_target_loc_and_dist,
                          dtypes=[string, np.float])

        Once NetworkBuilder.build() is called the 'connection_rule' will be called.
    """

    class ParamsRules(object):
        """A subclass to store indvidiual synpatic parameter rules"""
        def __init__(self, names, rule, rule_params, dtypes):
            self._names = names
            self._rule = rule
            self._rule_params = rule_params
            self._dtypes = self.__create_dtype_dict(names, dtypes)

        def __create_dtype_dict(self, names, dtypes):
            if isinstance(names, list):
                # TODO: compare size of names and dtypes
                return {n: dt for n, dt in zip(names, dtypes)}
            else:
                return {names: dtypes}

        @property
        def names(self):
            return self._names

        @property
        def rule(self):
            return connector.create(self._rule, **(self._rule_params or {}))

        @property
        def dtypes(self):
            return self._dtypes

        def get_prop_dtype(self, prop_name):
            return self._dtypes[prop_name]

    def __init__(self, sources=None, targets=None, connector=None, connector_params=None, iterator='one_to_one',
                 edge_type_properties=None):
        self._source_nodes = sources  # source nodes
        self._target_nodes = targets  # target nodes
        self._connector = connector  # function, list or value that determines connection between sources and targets
        self._connector_params = connector_params  # parameters passed into connector
        self._iterator = iterator  # rule for iterating between sources and targets
        self._edge_type_properties = edge_type_properties

        self._params = []
        self._param_keys = []

    @property
    def params(self):
        return self._params

    @property
    def source_nodes(self):
        return self._source_nodes

    @property
    def source_network_name(self):
        return self._source_nodes.network_name

    @property
    def target_nodes(self):
        return self._target_nodes

    @property
    def target_network_name(self):
        return self._target_nodes.network_name

    @property
    def connector(self):
        return self._connector

    @property
    def connector_params(self):
        return self._connector_params

    @property
    def iterator(self):
        return self._iterator

    @property
    def edge_type_properties(self):
        return self._edge_type_properties or {}

    @property
    def edge_type_id(self):
        # TODO: properly implement edge_type
        return self._edge_type_properties['edge_type_id']

    @property
    def property_names(self):
        if len(self._param_keys) == 0:
            return ['nsyns']
        else:
            return self._param_keys

    def properties_keys(self):
        ordered_keys = sorted(self.property_names)
        return str(ordered_keys)

    def max_connections(self):
        return len(self._source_nodes) * len(self._target_nodes)

    def add_properties(self, names, rule, rule_params=None, dtypes=None):
        """Add a synaptic property for an individual edge.

        Typically this requires a custom rule that will be used for every source/target pair of nodes and returns
        a value for the property. Such rule is a user defined function, and will be called by NetworkBuilder.build()::

            def syn_weight_by_loc(source, target):
                ...
                return syn_weight

            cm.add_properties('syn_weight', rule=syn_weight_by_loc, dtypes=np.float)

        If the connection_rule function requires additional parameters use the rule_params option::

            def syn_weight_by_loc(source, target, min_weight, max_weight):
                ...
                return syn_weight

            cm.add_properties('syn_weight',
                              rule=syn_weight_by_loc,
                              rule_params={'min_weigth':1.0e-7, 'max_weight': 5.0e-6},
                              dtypes=np.float)

        You may also have the rule define multiple connection parameters at one time::

            def conn_rule(source, target, min_weight, max_weight):
                ...
                return syn_weight, syn_location, delay

            cm.add_properties(['syn_weight', 'syn_loc', 'delay'],
                              rule=syn_weight_by_loc,
                              rule_params={'min_weigth':1.0e-7, 'max_weight': 5.0e-6},
                              dtypes=[np.float, str, np.float])

        :param names: list, or single string, of the property
        :param rule: function, list or value of property
        :param rule_params: when rule is a function, rule_params will be passed into function when called.
        :param dtypes: expected property type
        """
        self._params.append(self.ParamsRules(names, rule, rule_params, dtypes))
        self._param_keys += names

    def connection_itr(self):
        """Returns a generator that will iterate through the source/target pairs (as specified by the iterator function,
        and create a connection rule based on the connector.
        """
        conr = connector.create(self.connector, **(self.connector_params or {}))
        itr = iterator.create(self.iterator, conr, **({}))
        return itr(self.source_nodes, self.target_nodes, conr)



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import os
import glob
import csv
import pandas as pd

from bmtk.utils.reports.spike_trains import SpikeTrains, sort_order, sort_order_lu
from bmtk.simulator.pointnet.io_tools import io
from bmtk.simulator.pointnet.nest_utils import nest_version

import nest


try:
    MPI_RANK = nest.Rank()
    N_HOSTS = nest.NumProcesses()

except Exception as e:
    MPI_RANK = 0
    N_HOSTS = 1


def create_spike_detector_nest2(label):
    return nest.Create(
        "spike_detector", 1,
        {
            'label': label,
            'withtime': True,
            'withgid': True,
            'to_file': True
        }
    )


def create_spike_detector_nest3(label):
    return nest.Create(
        "spike_recorder", 1,
        {
            'label': label,
            'record_to': 'ascii'
        }
    )


def read_spikes_file_nest2(spike_trains_writer, gid_map, label):
    for gdf_path in glob.glob(label + '*.gdf'):
        with open(gdf_path, 'r') as csv_file:
            csv_reader = csv.reader(csv_file, delimiter='\t')
            for r in csv_reader:
                p = gid_map.get_pool_id(int(r[0]))
                spike_trains_writer.add_spike(node_id=p.node_id, timestamp=float(r[1]), population=p.population)


def read_spikes_file_nest3(spike_trains_writer, gid_map, label):
    for dat_path in glob.glob(label + '*.dat'):
        report_df = pd.read_csv(dat_path, index_col=False, sep='\t', comment='#')
        for nest_id, spikes_grp in report_df.groupby('sender'):
            gid = gid_map.get_pool_id(nest_id)
            timestamps = spikes_grp['time_ms'].values
            spike_trains_writer.add_spikes(node_ids=gid.node_id, timestamps=timestamps, population=gid.population)


if nest_version[0] >= 3:
    create_spike_detector = create_spike_detector_nest3
    read_spikes_file = read_spikes_file_nest3
    NEST_spikes_file_format = 'dat'
else:
    create_spike_detector = create_spike_detector_nest2
    read_spikes_file = read_spikes_file_nest2
    NEST_spikes_file_format = 'gdf'


class SpikesMod(object):
    """Module use for saving spikes

    """

    def __init__(self, tmp_dir, spikes_file_csv=None, spikes_file=None, spikes_file_nwb=None, spikes_sort_order=None,
                 cache_to_disk=True):
        def _get_path(file_name):
            # Unless file-name is an absolute path then it should be placed in the $OUTPUT_DIR
            if file_name is None:
                return None

            if os.path.isabs(file_name):
                return file_name
            else:
                abs_tmp = os.path.abspath(tmp_dir)
                abs_fname = os.path.abspath(file_name)
                if not abs_fname.startswith(abs_tmp):
                    return os.path.join(tmp_dir, file_name)
                else:
                    return file_name

        self._csv_fname = _get_path(spikes_file_csv)
        self._h5_fname = _get_path(spikes_file)
        self._nwb_fname = _get_path(spikes_file_nwb)

        self._tmp_dir = tmp_dir
        self._tmp_file_base = 'tmp_spike_times'
        self._spike_labels = os.path.join(self._tmp_dir, self._tmp_file_base)

        self._spike_writer = SpikeTrains(cache_dir=tmp_dir, cache_to_disk=cache_to_disk)
        self._spike_writer.delimiter = '\t'
        self._spike_writer.gid_col = 0
        self._spike_writer.time_col = 1
        self._sort_order = sort_order.none if not spikes_sort_order else sort_order_lu[spikes_sort_order]

        self._spike_detector = None

    def initialize(self, sim):
        self._spike_detector = create_spike_detector(self._spike_labels)
        nest.Connect(sim.net.gid_map.gids, self._spike_detector)

    def finalize(self, sim):
        # convert NEST gdf files into SONATA spikes/ format
        # TODO: Create a gdf_adaptor in bmtk/utils/reports/spike_trains to improve conversion speed.
        if MPI_RANK == 0:
            gid_map = sim.net.gid_map
            read_spikes_file(spike_trains_writer=self._spike_writer, gid_map=gid_map, label=self._spike_labels)
        io.barrier()

        if self._csv_fname is not None:
            self._spike_writer.to_csv(self._csv_fname, sort_order=self._sort_order)
            # io.barrier()

        if self._h5_fname is not None:
            # TODO: reimplement with pandas
            self._spike_writer.to_sonata(self._h5_fname, sort_order=self._sort_order)
            # io.barrier()

        if self._nwb_fname is not None:
            self._spike_writer.to_nwb(self._nwb_fname, sort_order=self._sort_order)
            # io.barrier()

        self._spike_writer.close()
        self._clean_files()

    def _clean_files(self):
        if MPI_RANK == 0:
            for nest_file in glob.glob(self._spike_labels + '*.' + NEST_spikes_file_format):
                os.remove(nest_file)



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
class EdgeSet(object):
    def __init__(self, edge_ids, population):
        self._edge_ids = edge_ids
        self._population = population
        self._n_edges = len(self._edge_ids)
        self.__itr = 0

    def __iter__(self):
        self.__itr = 0
        return self

    def next(self):
        if self.__itr >= self._n_edges:
            raise StopIteration

        next_edge = self._population.iloc(self._edge_ids[self.__itr])
        self.__itr += 1
        return next_edge


class Edge(object):
    def __init__(self, src_node_id, trg_node_id, source_pop, target_pop, group_id, group_props, edge_types_props):
        self._src_node_id = src_node_id
        self._trg_node_id = trg_node_id
        self._source_population = source_pop
        self._target_population = target_pop
        self._group_props = group_props
        self._group_id = group_id
        self._edge_type_props = edge_types_props

    @property
    def source_node_id(self):
        return self._src_node_id

    @property
    def target_node_id(self):
        return self._trg_node_id

    @property
    def source_population(self):
        return self._source_population

    @property
    def target_population(self):
        return self._target_population

    @property
    def group_id(self):
        return self._group_id

    @property
    def edge_type_id(self):
        return self._edge_type_props['edge_type_id']

    @property
    def dynamics_params(self):
        raise NotImplementedError

    def __getitem__(self, prop_key):
        if prop_key in self._group_props:
            return self._group_props[prop_key]
        elif prop_key in self._edge_type_props:
            return self._edge_type_props[prop_key]
        else:
            raise KeyError('Property {} not found in edge.'.format(prop_key))

    def __contains__(self, prop_key):
        return prop_key in self._group_props or prop_key in self._edge_type_props


import logging
from optparse import OptionParser

from .create_environment import create_environment


logger = logging.getLogger(__name__)


def __list_parser(option, opt, value, parser):
    """Helper function for parsing inputs, since they may contain a list of values that are comma (,) separated.

    eg. ---spike-trains=lgn:inputs/lgn_spikes.h5,bkg:inputs/bkg_spikes.h5
    """
    # parser.values.has_membrane_report = True
    if ',' in value:
        try:
            setattr(parser.values, option.dest, [float(v) for v in value.split(',')])
        except ValueError as ve:
            setattr(parser.values, option.dest, value.split(','))

    else:
        setattr(parser.values, option.dest, value)


def __split_list(options, opt_name):
    opt_vals = getattr(options, opt_name)
    updated_vals = []
    if opt_vals is not None:
        opt_strings = [opt_vals] if isinstance(opt_vals, str) else list(opt_vals)
        for opt_str in opt_strings:
            vals = opt_str.split(':')
            if len(vals) == 1:
                updated_vals.append((None, vals[0]))
            elif len(vals) == 2:
                updated_vals.append((vals[0], vals[1]))
            else:
                parser.error('Cannot parse spike-input string <pop1>:<spikes-file1>,<pop2>:<spikes-file2>,...')

        setattr(options, opt_name, updated_vals)


if __name__ == '__main__':
    # logging.basicConfig(level=logging.INFO, format='%(module)s [%(levelname)s] %(message)s')
    logging.basicConfig(level=logging.WARN, format='bmtk.utils.create_environment [%(levelname)s] %(message)s')

    parser = OptionParser(usage="Usage: python %prog [options] [bionet|pointnet|popnet|filternet] sim_dir")
    parser.add_option('-n', '--network-dir', dest='network_dir', default=None,
                      help="Use an exsting directory with network files.")
    parser.add_option('-o', '--output-dir', dest='output_dir', default=None,
                      help="Directory to use for storing output of simulations.")
    parser.add_option('-c', '--components-dir', dest='components_dir', default=None,
                      help="Directory to use for parameter files, morphology, and other used components.")
    parser.add_option('--tstop', type='float', dest='tstop', default=1000.0)
    parser.add_option('--dt', type=float, dest='dt', help='simulation time step dt', default=0.001)
    parser.add_option('--report-vars', dest='report_vars', type='string', action='callback',
                      callback=__list_parser, default=[],
                      help='A list of membrane variables to record from; v, cai, etc.')
    parser.add_option('--report-nodes', dest='report_nodes', type='string', action='callback',
                      callback=__list_parser, default=None)
    parser.add_option('--iclamp', dest='current_clamp', type='string', action='callback',
                      callback=__list_parser, default=None,
                      help='Adds a soma current clamp using three variables: <amp>,<delay>,<duration> (nA, ms, ms)')
    parser.add_option('--spikes-inputs', dest='spikes_inputs', type='string', action='callback', callback=__list_parser,
                      default=None,
                      help='Spike trains to use for inputs stimulations, a list in the form of ' 
                           '<pop1>:<spikes_file_1.h5>[,<pop2>:<spikes_file_1.h5>,...]')
    parser.add_option('--rates-inputs', dest='rates_inputs', type='string', action='callback', callback=__list_parser,
                      default=None,
                      help='Population rates file to use for inputs stimulations, a list in the form of ' 
                           '<pop1>:<rates_file_1.h5>[,<pop2>:<ratess_file_1.h5>,...]')
    parser.add_option('--include-examples', dest='include_examples', action='store_true', default=False,
                      help='Copies component files used by examples and tutorials.')
    parser.add_option('--compile-mechanisms', dest='compile_mechanisms', action='store_true', default=False,
                      help='Will try to compile the NEURON mechanisms (BioNet only).')
    parser.add_option('--config-file', dest='config_file', type='string', default=None,
                      help='File name of conguration json file.')
    parser.add_option('--config-name', dest='config_name', type='string', default=None,
                      help='General name of environment configuration file(s).')
    parser.add_option('--split-configs', dest='split_configs', action='store_true', default=False,
                      help='If option is used then the configuration files will be split into different circuit and '
                      'simulation configuration files')
    parser.add_option('--network-filter', dest='network_filter', type='string', action='callback',
                      callback=__list_parser, default=None,
                      help='Option to filter which files from --network-dir are included in configuration, A list of '
                      'files, v1_node.h5,v1_node_types.csv or a string v1,lgn.')
    parser.add_option('--run-script', dest='run_script', type='string', default=None,
                      help='Name of bmtk simulation run *.py script used to run the simulation. By default name will '
                      'be run_{simulator}.py')
    parser.add_option('--overwrite', dest='overwrite', action='store_true', default=False,
                      help='Overwrite existing configuration/run scripts if exists.')

    options, args = parser.parse_args()

    # Check the passed in argments are correct. [sim] </path/to/dir/>
    if len(args) < 2:
        parser.error('Invalid number of arguments, Please specify a target simulation (bionet, pointnet, filternet,'
                     'popnet) and the path to save the simulation environment.')
        exit(1)
    elif len(args) > 2:
        parser.error('Unrecognized arguments {}'.format(args[2:]))
        exit(1)
    else:
        simulator = args[0].lower()
        if simulator not in ['bionet', 'popnet', 'pointnet', 'filternet']:
            parser.error('Must specify one target simulator. options: "bionet", pointnet", "popnet", "filternet"')
        base_dir = args[1]

    if options.current_clamp is not None:
        cc_args = options.current_clamp
        if len(cc_args) != 3:
            parser.error('Invalid arguments for current clamp, requires three floating point numbers '
                         '<ampage>,<delay>,<duration> (nA, ms, ms)')
        iclamp_args = {'amp': float(cc_args[0]), 'delay': float(cc_args[1]), 'duration': float(cc_args[2])}
        options.current_clamp = iclamp_args
    else:
        options.current_clamp = None

    __split_list(options, 'spikes_inputs')
    __split_list(options, 'report_vars')
    __split_list(options, 'rates_inputs')

    if options.run_script is None:
        options.run_script = True

    create_environment(simulator=simulator, base_dir=base_dir, **vars(options))



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import os
import json
import functools
import nest

from six import string_types
import numpy as np

from bmtk.simulator.core.simulator_network import SimNetwork
from bmtk.simulator.pointnet.sonata_adaptors import PointNodeAdaptor, PointEdgeAdaptor
from bmtk.simulator.pointnet import pyfunction_cache
from bmtk.simulator.pointnet.io_tools import io
from bmtk.simulator.pointnet.nest_utils import nest_version
from .gids import GidPool


def set_spikes_nest2(node_id, nest_obj, spike_trains):
    st = spike_trains.get_times(node_id=node_id)
    if st is None or len(st) == 0:
        return

    st = np.array(st)
    if np.any(st <= 0.0):
        # NRN will fail if VecStim contains negative spike-time, throw an exception and log info for user
        io.log_exception('spike train {} contains negative/zero time, unable to run virtual cell in NEST'.format(st))
    st.sort()
    nest.SetStatus([nest_obj], {'spike_times': st})


def set_spikes_nest3(node_id, nest_obj, spike_trains):
    st = spike_trains.get_times(node_id=node_id)
    if st is None or len(st) == 0:
        return

    st = np.array(st)
    if np.any(st <= 0.0):
        io.log_exception('spike train {} contains negative/zero time, unable to run virtual cell in NEST'.format(st))
    st.sort()
    nest.SetStatus(nest_obj, {'spike_times': st})


if nest_version[0] >= 3:
    set_spikes = set_spikes_nest3
else:
    set_spikes = set_spikes_nest2


class PointNetwork(SimNetwork):
    def __init__(self, **properties):
        super(PointNetwork, self).__init__(**properties)
        self._io = io

        self.__weight_functions = {}
        self._params_cache = {}

        self._virtual_ids_map = {}

        self._batch_nodes = True

        # self._nest_id_map = {}
        self._nestid2nodeid_map = {}

        self._nestid2gid = {}

        self._nodes_table = {}
        self._gid2nestid = {}

        self._gid_map = GidPool()
        self._virtual_gids = GidPool()

    @property
    def py_function_caches(self):
        return pyfunction_cache

    @property
    def gid_map(self):
        return self._gid_map

    @property
    def gid_pool(self):
        return self._gid_map

    def get_nodes_df(self, population):
        nodes_adaptor = self.get_node_population(population)
        return nodes_adaptor.nodes_df()

    def __get_params(self, node_params):
        if node_params.with_dynamics_params:
            # TODO: use property, not name
            return node_params['dynamics_params']

        params_file = node_params[self._params_column]
        # params_file = self._MT.params_column(node_params) #node_params['dynamics_params']
        if params_file in self._params_cache:
            return self._params_cache[params_file]
        else:
            params_dir = self.get_component('models_dir')
            params_path = os.path.join(params_dir, params_file)
            params_dict = json.load(open(params_path, 'r'))
            self._params_cache[params_file] = params_dict
            return params_dict

    def _register_adaptors(self):
        super(PointNetwork, self)._register_adaptors()
        self._node_adaptors['sonata'] = PointNodeAdaptor
        self._edge_adaptors['sonata'] = PointEdgeAdaptor

    # TODO: reimplement with py_modules like in bionet
    def add_weight_function(self, fnc, name=None, **kwargs):
        fnc_name = name if name is not None else function.__name__
        self.__weight_functions[fnc_name] = functools.partial(fnc)

    def set_default_weight_function(self, fnc):
        self.add_weight_function(fnc, 'default_weight_fnc', overwrite=True)

    def get_weight_function(self, name):
        return self.__weight_functions[name]

    def get_node_id(self, population, node_id):
        pop = self.get_node_population(population)
        return pop.get_node(node_id)

    def build_nodes(self):
        for node_pop in self.node_populations:
            pop_name = node_pop.name
            gid_map = self.gid_map

            gid_map.create_pool(pop_name)
            if node_pop.internal_nodes_only:
                for node in node_pop.get_nodes():
                    node.build()
                    gid_map.add_nestids(name=pop_name, node_ids=node.node_ids, nest_ids=node.nest_ids)

            elif node_pop.mixed_nodes:
                for node in node_pop.get_nodes():
                    if node.model_type != 'virtual':
                        node.build()
                        gid_map.add_nestids(name=pop_name, node_ids=node.node_ids, nest_ids=node.nest_ids)

    def build_recurrent_edges(self, force_resolution=False):
        recurrent_edge_pops = [ep for ep in self._edge_populations if not ep.virtual_connections]
        if not recurrent_edge_pops:
            return

        for edge_pop in recurrent_edge_pops:
            for edge in edge_pop.get_edges():
                nest_srcs = self.gid_map.get_nestids(edge_pop.source_nodes, edge.source_node_ids)
                nest_trgs = self.gid_map.get_nestids(edge_pop.target_nodes, edge.target_node_ids)
                self._nest_connect(nest_srcs, nest_trgs, conn_spec='one_to_one', syn_spec=edge.nest_params)

    def find_edges(self, source_nodes=None, target_nodes=None):
        # TODO: Move to parent
        selected_edges = self._edge_populations[:]

        if source_nodes is not None:
            selected_edges = [edge_pop for edge_pop in selected_edges if edge_pop.source_nodes == source_nodes]

        if target_nodes is not None:
            selected_edges = [edge_pop for edge_pop in selected_edges if edge_pop.target_nodes == target_nodes]

        return selected_edges

    def add_spike_trains(self, spike_trains, node_set, sg_params={'precise_times': True}):
        # Build the virtual nodes
        src_nodes = [node_pop for node_pop in self.node_populations if node_pop.name in node_set.population_names()]
        virt_gid_map = self._virtual_gids
        for node_pop in src_nodes:
            if node_pop.name in self._virtual_ids_map:
                continue

            virt_node_map = {}
            if node_pop.virtual_nodes_only:
                for node in node_pop.get_nodes():
                    nest_objs = nest.Create('spike_generator', node.n_nodes, sg_params)
                    nest_ids = nest_objs.tolist() if nest_version[0] >= 3 else nest_objs

                    virt_gid_map.add_nestids(name=node_pop.name, nest_ids=nest_ids, node_ids=node.node_ids)
                    for node_id, nest_obj, nest_id in zip(node.node_ids, nest_objs, nest_ids):
                        virt_node_map[node_id] = nest_id
                        set_spikes(node_id=node_id, nest_obj=nest_obj, spike_trains=spike_trains)

            elif node_pop.mixed_nodes:
                for node in node_pop.get_nodes():
                    if node.model_type != 'virtual':
                        continue

                    nest_ids = nest.Create('spike_generator', node.n_nodes, sg_params)
                    for node_id, nest_id in zip(node.node_ids, nest_ids):
                        virt_node_map[node_id] = nest_id
                        set_spikes(node_id=node_id, nest_id=nest_id, spike_trains=spike_trains)

            self._virtual_ids_map[node_pop.name] = virt_node_map

        # Create virtual synaptic connections
        for source_reader in src_nodes:
            for edge_pop in self.find_edges(source_nodes=source_reader.name):
                for edge in edge_pop.get_edges():
                    nest_trgs = self.gid_map.get_nestids(edge_pop.target_nodes, edge.target_node_ids)
                    nest_srcs = virt_gid_map.get_nestids(edge_pop.source_nodes, edge.source_node_ids)
                    self._nest_connect(nest_srcs, nest_trgs, conn_spec='one_to_one', syn_spec=edge.nest_params)

    def _nest_connect(self, nest_srcs, nest_trgs, conn_spec='one_to_one', syn_spec=None):
        """Calls nest.Connect but with some extra error logging and exception handling."""
        try:
            nest.Connect(nest_srcs, nest_trgs, conn_spec=conn_spec, syn_spec=syn_spec)

        except nest.kernel.NESTErrors.BadDelay as bde:
            # An occuring issue is when dt > delay, add some extra messaging in log to help users fix problem.
            res_kernel = nest.GetKernelStatus().get('resolution', 'NaN')
            delay_edges = syn_spec.get('delay', 'NaN')
            msg = 'synaptic "delay" value in edges ({}) is not compatable with simulator resolution/"dt" ({})'.format(
                delay_edges, res_kernel
            )
            self.io.log_error('{}{}'.format(bde.errorname, bde.errormessage))
            self.io.log_error(msg)
            raise

        except Exception as e:
            # Record exception to log file.
            self.io.log_error(str(e))
            raise



import numpy as np
import hashlib
import logging


try:
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
    mpi_rank = comm.Get_rank()
    mpi_size = comm.Get_size()
    barrier = comm.barrier

except ImportError:
    mpi_rank = 0
    mpi_size = 1
    barrier = lambda: None


MPI_fail_params_nonuniform = True  # throw exception unless params across MPI ranks are the same.
logger = logging.getLogger(__name__)


def add_hdf5_attrs(hdf5_handle):
    # TODO: move this as a utility function
    hdf5_handle['/'].attrs['magic'] = np.uint32(0x0A7A)
    hdf5_handle['/'].attrs['version'] = [np.uint32(0), np.uint32(1)]


def list_to_hash(str_list):
    str_list = str_list.copy()
    str_list.sort()
    combined_keys = ':'.join(str_list).encode('utf-8')
    return hashlib.md5(combined_keys).hexdigest()


def check_properties_across_ranks(properties, graph_type='node'):
    """Checks that a properties table is consistant across all MPI ranks. Mainly used by add_nodes() and add_edges()
    method due to bug where using random generator without rng_seed was causing issues building the network properties
    consistantly using multiple cores.

    Will throw an Exception or a warning message (if MPI_fail_params_nonuniform is false)

    :param properties: A dictionary
    :param graph_type: 'node' or 'edge', used in error message. default 'node'.
    """
    if mpi_size < 2:
        return

    # Check that model_properties have the same number of items and the keys match
    n_args = len(properties)
    ranked_args = comm.allgather(n_args)
    if len(set(ranked_args)) > 1:
        err_msg = '{} properties are not the same across all ranks.'.format(graph_type)
        if not MPI_fail_params_nonuniform:
            logger.warning(err_msg)
        else:
            raise IndexError(err_msg)

    if n_args == 0:
        return

    # create a string/id that will be uniform across all ranks, even if the dict on one rank returns keys out-of-order.
    prop_keys = list(properties.keys())
    prop_keys.sort()
    combined_keys = ':'.join(prop_keys).encode('utf-8')
    hash_id = hashlib.md5(combined_keys).hexdigest()
    ranked_keys = comm.allgather(hash_id)
    if len(set(ranked_keys)) > 1:
        err_msg = '{} properties are not the same across all ranks.'.format(graph_type)
        if not MPI_fail_params_nonuniform:
            logger.warning(err_msg)
        else:
            raise IndexError(err_msg)

    # For each item in model_properties dictionary try to check that values are the same
    for pkey in prop_keys:
        # Don't use Dict.items() method since it is possible the ret order is different across ranks.
        pval = properties[pkey]

        try:
            if isinstance(pval, bytes):
                phash = hashlib.md5(pval).hexdigest()

            elif isinstance(pval, str):
                phash = hashlib.md5(pval.encode('utf-8')).hexdigest()

            elif isinstance(pval, (int, float, bool)):
                phash = pval

            elif isinstance(pval, (list, tuple)):
                joined_keys = ':'.join([str(p) for p in pval]).encode('utf-8')
                phash = hashlib.md5(joined_keys).hexdigest()

            elif isinstance(pval, np.ndarray):
                phash = hashlib.md5(pval.data.tobytes()).hexdigest()

            else:
                continue

        except TypeError as te:
            # If the hashing fails assume there is no MPI data issue and continue with the next property.
            continue

        ranked_vals = comm.allgather(phash)
        if len(set(ranked_vals)) > 1:
            err_msg = '{} property "{}" varies across ranks, please make sure parameter value is uniform across all' \
                      'ranks or set bmtk.builder.MPI_fail_params_nonuniform to False'.format(graph_type, pkey)
            if not MPI_fail_params_nonuniform:
                logger.warning(err_msg)
            else:
                raise TypeError(err_msg)
